from .entities import (
    Entity,
    Class,
    Individual,
    DataProperty,
    ObjectProperty,
    DataFieldType,
    Datatype,
    DatatypeType,
)
from .relation import TargetStatement, StatementType, RestrictionType, Relation
from .combined_relations import (
    CombinedRelation,
    CombinedObjectRelation,
    CombinedDataRelation,
)
from .source import DependencyStatement, ParsingError, Source
from .vocabulary import IdType, LabelSummary, VocabularySettings, Vocabulary



"""Vocabulary Models for Relations"""

from typing import Set, Dict, TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from . import Vocabulary

from aenum import Enum
from typing import List, TYPE_CHECKING

from pydantic import BaseModel, Field

from .source import DependencyStatement

if TYPE_CHECKING:
    from . import Vocabulary, IdType


class StatementType(str, Enum):
    """
    A statement is either a leaf and holds an iri/label or it is a combination
    of leafs with or / and
    """

    OR = "or"
    AND = "and"
    LEAF = "leaf"


class TargetStatement(BaseModel):
    """
    A target statement is the statement the sits in a relation statement behind
    the restrictionType:
    E.g: consists_of some Device or Sensor.
    here Device or Sensor is the targetstatement as it sits behind "some"

    A targetstatement is build recursively: It is either a Leaf: str or a union
    (or) or an intersection(and) of targetstatements

    the combination of statements is purely logical and not numerical: device
        and device is true as soon as one device is given,
        it does not need two separate devices.
    """

    target_data_value: Optional[str] = Field(
        default=None,
        description="Holds the value if the relation is a hasValue (LEAF only)",
    )
    target_iri: str = Field(default="", description="The IRI of the target (LEAF only)")
    target_statements: List["TargetStatement"] = Field(
        default=[],
        description="The targetstatements that are combined with this "
        "targetstatement (and/or only)",
    )
    type: StatementType = Field(
        default=StatementType.LEAF, description="Statement types"
    )

    def set_target(self, target_iri: str, target_data_value: str = None):
        """Set target for this statement and make it a LEAF statement

        Args:
            target_iri (str): iri of the target (class or datatype iri)
            target_data_value (str): Value of the targetstatment if no IRI but
             a fixed default value is given

        Returns:
            None
        """
        self.type = StatementType.LEAF
        self.target_iri = target_iri
        self.target_data_value = target_data_value

    def get_all_targets(self) -> List[List[str]]:
        """Extract possible targets out of statements
        interpretation: [[a,b], [c]]-> (a and b) or c -> the target needs to
        have ancestors(or be): (both a anb b) or c

        items inside the inner brackets are connected via and the innerbrackets
        are all connect over or

        Returns:
            List[List[str]]
        """
        if self.type == StatementType.LEAF:
            if self.target_data_value is not None:
                return [[]]
            else:
                return [[self.target_iri]]
        else:
            collection = []  # form: [ [[]], [[],[]], ..]
            for statement in self.target_statements:
                collection.append(statement.get_all_targets())

            if self.type == StatementType.OR:
                result = []
                for sublist in collection:  # sublist form: [[],[],...]
                    result.extend(sublist)
                return result

            else:  # AND
                # with and we distribute our lists
                # example: col= [ [[1],[2]], [[3,4]] ] => [[1,3,4], [2,3,4]]
                # we build the results in an all in one way, we compute the
                # number of entries in the final solution
                # for each list li we fill the i-th position of all results

                # example: col= [ [[a,b]], [[c],[d]] , [[e],[f]] ] =>
                # statement: (a and b) and (c or d) and (e or f)
                # 0 : res = [[], [], [], []]
                # 1 : res = [[a,b], [a,b], [a,b], [a,b]]
                # 2 : res = [[a,b,c], [a,b,c], [a,b,d], [a,b,d]]
                # 3 : res = [[a,b,c,e], [a,b,c,f], [a,b,d,e], [a,b,d,f]]

                result = []  # result form: [[],[],...]
                lengths = []
                number_of_entries = 1
                for sublist in collection:  # sublist form: [[],[],...]
                    number_of_entries = number_of_entries * len(sublist)
                    lengths.append(len(sublist))

                # init with empty lists
                for empty in range(number_of_entries):
                    result.append([])

                for i in range(0, len(collection)):
                    mod = 1
                    for j in range(i + 1, len(lengths)):
                        mod = mod * lengths[j]

                    counter = 0
                    while counter < number_of_entries:
                        for sublist in collection[i]:
                            for entry in sublist:
                                for j in range(mod):
                                    result[counter].append(entry)
                                    counter += 1
                return result

    def to_string(self, vocabulary: "Vocabulary") -> str:
        """Get a string representation of the targetstatment

        Args:
            vocabulary (Vocabulary): vocabulary of the project

        Returns:
            str
        """

        if self.type == StatementType.LEAF:
            label = self.retrieve_label(vocabulary)
            if label == "":
                return self.target_iri
            return label
        else:
            result = "(" + self.target_statements[0].to_string(vocabulary)
            for statement in self.target_statements[1:]:
                result += " " + self.type + " "
                result += statement.to_string(vocabulary) + ")"

            return result

    def is_fulfilled_by_iri_value(self, value: str, ancestor_values: List[str]) -> bool:
        """
        Test if a set of values fulfills the targetstatement;
        Only for objectRelations

        Args:
            value (str): value to check: Class_iri of instance/individual
            ancestor_values(List[List[str]]): List containing the ancestors iris
                for each value (linked over index)
        Returns:
            bool
        """

        targets = self.get_all_targets()

        values = ancestor_values
        values.append(value)

        # one sublist of targets needs to be fulfilled targets:
        # [(a and b) or c or (d and a),....]
        for sublist in targets:
            sublist_fulfilled = True
            for item in sublist:
                if item not in values:
                    sublist_fulfilled = False

            if sublist_fulfilled:
                return True
        return False

    def is_fulfilled_by_data_value(self, value: str, vocabulary: "Vocabulary") -> bool:
        """
        Test if a set of values fulfills the targetstatement;
        Only for dataRelations

        Args:
            value (List[str]):  value to check
            vocabulary (Vocabulary)
        Returns:
            bool
        """

        # a data target_statement theoraticly only has one statement
        if self.target_data_value is not None:
            return value == self.target_data_value

        from .vocabulary import IdType

        if not vocabulary.is_id_of_type(self.target_iri, IdType.datatype):
            return False
        else:
            datatype = vocabulary.get_datatype(self.target_iri)
            return datatype.value_is_valid(value)

    def retrieve_label(self, vocabulary: "Vocabulary") -> str:
        """Get the label of the target_iri. Only logical for Leaf statements

        Args:
            vocabulary (Vocabulary): Vocabulary of the project

        Returns:
            str
        """
        if self.type == StatementType.LEAF:
            if self.target_data_value is not None:
                return self.target_data_value
            else:
                return vocabulary.get_label_for_entity_iri(self.target_iri)
        return ""

    def get_dependency_statements(
        self, vocabulary: "Vocabulary", ontology_iri: str, class_iri: str
    ) -> List[DependencyStatement]:
        """
        Get a list of all pointers/iris that are not contained in the
        vocabulary. Purging is done in class

        Args:
            vocabulary (Vocabulary): Vocabulary of this project
            ontology_iri (str): IRI of the source ontology
            class_iri (str): IRI of class (legacy)

        Returns:
            List[Dict[str, str]]: List of purged statements dicts with keys:
            Parent Class, class, dependency, fulfilled
        """
        statements = []
        if self.type == StatementType.LEAF:

            # if we have a given data value, we do not have an iri
            if self.target_data_value is None:
                # check if predefined datatype
                if not vocabulary.iri_is_predefined_datatype(self.target_iri):
                    found = (
                        self.target_iri in vocabulary.classes
                        or self.target_iri in vocabulary.datatypes
                        or self.target_iri in vocabulary.individuals
                    )
                    statements.append(
                        DependencyStatement(
                            type="Relation Target",
                            class_iri=class_iri,
                            dependency_iri=self.target_iri,
                            fulfilled=found,
                        )
                    )
        else:
            for target_statement in self.target_statements:
                statements.extend(
                    target_statement.get_dependency_statements(
                        vocabulary, ontology_iri, class_iri
                    )
                )
        return statements


# target_statements is a forward reference, so that the class can refer to
# itself this forward reference need to be resolved after the class has fully
# loaded
TargetStatement.model_rebuild()


class RestrictionType(str, Enum):
    """RestrictionTypes, as defined for OWL"""

    _init_ = "value __doc__"

    some = "some", "at least 1 value of that target"
    only = "only", "only value of that target"
    min = "min", "min n values of that target"
    max = "max", "max n values of that target"
    exactly = "exactly", "exactly n values of that target"
    value = "value", "predefined value"


class Relation(BaseModel):
    """
    A Relation is defined in the source for a class.
    It has the form: RestrictionType property target_statement

    It defines a set of allowed/required values which each instance of this
    class can/should have under this property

    A Relation is defined in a OWL:class, but all child classes of that class
    inherit it
    """

    id: str = Field(description="Unique generated Relation ID, " "for internal use")
    restriction_type: RestrictionType = Field(
        default=None, description="Restriction type of this relation"
    )
    restriction_cardinality: int = Field(
        default=-1, description="Only needed for min, max, equaly states the 'n'"
    )
    property_iri: str = Field(
        default="", description="IRI of the property (data- or object-)"
    )
    target_statement: TargetStatement = Field(
        default=None,
        description="Complex statement which classes/datatype_catalogue "
        "are allowed/required",
    )

    def get_targets(self) -> List[List[str]]:
        """Get all targets specified in the target statement in AND-OR Notation

        Returns:
            List[List[str]]: [[a,b],[c]] either a and b needs to present, or c
        """
        return self.target_statement.get_all_targets()

    def to_string(self, vocabulary: "Vocabulary") -> str:
        """Get a string representation of the relation

        Args:
            vocabulary (Vocabulary): Vocabulary of this project

        Returns:
            str
        """

        if self.restriction_cardinality == -1:
            return "{} {}".format(
                self.restriction_type, self.target_statement.to_string(vocabulary)
            )
        else:
            return (
                self.restriction_type
                + " "
                + str(self.restriction_cardinality)
                + " "
                + self.target_statement.to_string(vocabulary)
            )

    def is_restriction_fulfilled(
        self, number_of_fulfilling_values: int, total_number_of_values: int
    ) -> bool:
        """Test if the restriction type is fulfilled by comparing the number of
        fulfilling values against the total
        number of values given

        Args:
            number_of_fulfilling_values (int): Number of values that fulfill the
                relation
            total_number_of_values (int): the number of values given for this
                relation

        Returns:
            bool
        """

        if self.restriction_type == RestrictionType.some:
            return number_of_fulfilling_values >= 1
        if self.restriction_type == RestrictionType.only:
            return number_of_fulfilling_values == total_number_of_values
        if self.restriction_type == RestrictionType.min:
            return number_of_fulfilling_values >= (int)(self.restriction_cardinality)
        if self.restriction_type == RestrictionType.max:
            return number_of_fulfilling_values <= (int)(self.restriction_cardinality)
        if self.restriction_type == RestrictionType.exactly:
            return number_of_fulfilling_values == (int)(self.restriction_cardinality)
        if self.restriction_type == RestrictionType.value:
            return number_of_fulfilling_values >= 1

    def get_dependency_statements(
        self, vocabulary: "Vocabulary", ontology_iri: str, class_iri: str
    ) -> List[DependencyStatement]:
        """Get a list of all pointers/iris that are not contained in the
            vocabulary
            Purging is done in class

        Args:
            vocabulary (Vocabulary): Vocabulary of this project
            ontology_iri (str): IRI of the source ontology
            class_iri (str): IRI of class (legacy)

        Returns:
            List[Dict[str, str]]: List of purged statements dicts with keys:
                Parent Class, class, dependency, fulfilled
        """
        statements = []

        found = (
            self.property_iri in vocabulary.object_properties
            or self.property_iri in vocabulary.data_properties
        )

        statements.append(
            DependencyStatement(
                type="Relation Property",
                class_iri=class_iri,
                dependency_iri=self.property_iri,
                fulfilled=found,
            )
        )

        statements.extend(
            self.target_statement.get_dependency_statements(
                vocabulary, ontology_iri, class_iri
            )
        )

        return statements

    def is_fulfilled_with_iris(
        self,
        vocabulary: "Vocabulary",
        values: List[str],
        ancestor_values: List[List[str]],
    ) -> bool:
        """Test if a set of values fulfills the rules of the relation

        Args:
            vocabulary (Vocabulary): Vocabulary of the project
            values (List[str]):  List of values to check
            ancestor_values(List[List[str]]): List containing the ancestors iris
                for each value (linked over index)
        Returns:
            bool
        """
        number_of_fulfilling_values = 0
        for i in range(len(values)):
            if self.target_statement.is_fulfilled_by_iri_value(
                values[i], ancestor_values[i]
            ):
                number_of_fulfilling_values += 1

        return self.is_restriction_fulfilled(number_of_fulfilling_values, len(values))

    def is_fulfilled_with_values(
        self, vocabulary: "Vocabulary", values: List[str]
    ) -> bool:
        """Test if a set of values fulfills the rules of the relation.
        Used if property is a data property

        Args:
            vocabulary (Vocabulary): Vocabulary of the project
            values (List[str]):  List of values to check

        Returns:
            bool
        """
        number_of_fulfilling_values = 0

        for i in range(len(values)):
            if self.target_statement.is_fulfilled_by_data_value(values[i], vocabulary):
                number_of_fulfilling_values += 1

        return self.is_restriction_fulfilled(number_of_fulfilling_values, len(values))

    def get_all_possible_target_class_iris(self, vocabulary: "Vocabulary") -> Set[str]:
        """Get a set of class iris that are possible values for an
        objectRelation

        Args:
            vocabulary (Vocabulary): Vocabulary of this project

        Returns:
            Set[str]: class_iris
        """

        # if the relation is of type value it only defines that this relation
        # has the given values
        # not that it could have some more
        if self.restriction_type == RestrictionType.value:
            return set()

        possible_class_iris = set()
        targets = self.get_targets()
        for target_list in targets:
            for class_ in vocabulary.get_classes():
                if class_.is_child_of_all_classes(target_list):
                    possible_class_iris.add(class_.iri)
                    children = vocabulary.get_class_by_iri(class_.iri).child_class_iris
                    possible_class_iris.update(children)

        return possible_class_iris

    def get_possible_enum_target_values(self, vocabulary: "Vocabulary") -> List[str]:
        """Get all allowed enum target values for a data relation

        Args:
            vocabulary (Vocabulary): Vocabulary of this project

        Returns:
            List[str]
        """
        targets: List[List[str]] = self.target_statement.get_all_targets()

        from .vocabulary import IdType

        # methode only makes sense for data relations
        if not vocabulary.is_id_of_type(self.property_iri, IdType.data_property):
            return []

        res = []
        # as it is a datarelation the targets should only contain single lists,
        # to be flexible with changes we loop also
        # these lists
        for list in targets:
            for entry_iri in list:
                if vocabulary.is_id_of_type(entry_iri, IdType.datatype):
                    datatype = vocabulary.get_datatype(entry_iri)
                    res.extend(datatype.enum_values)

        return res

    def get_all_target_iris(self) -> Set[str]:
        """Get all iris of targets

        Returns:
            Set(str)
        """
        iris = set()

        statements = [self.target_statement]

        while len(statements) > 0:
            statement = statements.pop()
            if statement.type == StatementType.LEAF:
                if not statement.target_iri == "":
                    iris.add(statement.target_iri)
            else:
                statements.extend(statement.target_statements)

        return iris

    def export_rule(self, vocabulary: "Vocabulary") -> (str, str):
        """Get the rule as string

        Args:
           vocabulary (Vocabulary): Vocabulary of the project

        Returns:
           str
        """
        targets = []
        for inner_list in self.get_targets():
            new_list = []
            targets.append(new_list)
            for iri in inner_list:
                new_list.append(vocabulary.get_label_for_entity_iri(iri))

        if (int)(self.restriction_cardinality) > 0:
            return (
                f'"{self.restriction_type.value}|' f'{self.restriction_cardinality}"',
                targets,
            )
        else:
            return f'"{self.restriction_type.value}"', targets



"""
Functions to clean up a tenant within a fiware based platform.
"""

import warnings
from functools import wraps

from pydantic import AnyHttpUrl, AnyUrl
from requests import RequestException
from typing import Callable, List, Union
from filip.models import FiwareHeader, FiwareLDHeader
from filip.clients.ngsi_v2 import ContextBrokerClient, IoTAClient, QuantumLeapClient
from filip.clients.ngsi_ld.cb import ContextBrokerLDClient
from filip.models.ngsi_ld.context import ActionTypeLD
import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def clear_context_broker_ld(
    url: str = None,
    fiware_ld_header: FiwareLDHeader = None,
    cb_ld_client: ContextBrokerLDClient = None,
):
    """
    Function deletes all entities and subscriptions for a tenant in an LD context broker.

    Args:
        url: Url of the context broker LD
        fiware_ld_header: header of the NGSI-LD tenant
        cb_ld_client: NGSI-LD context broker client object

    Returns:

    """
    assert url or cb_ld_client, "Either url or client object must be given"
    # create client
    if cb_ld_client is None:
        client = ContextBrokerLDClient(url=url, fiware_header=fiware_ld_header)
    else:
        client = cb_ld_client
    # clean entities iteratively
    try:
        entity_list = True
        while entity_list:
            entity_list = client.get_entity_list(limit=100)
            if entity_list:
                client.entity_batch_operation(
                    action_type=ActionTypeLD.DELETE, entities=entity_list
                )
    except RequestException as e:
        logger.warning("Could not clean entities completely")
        raise

    # clean subscriptions
    try:
        sub_list = cb_ld_client.get_subscription_list()
        for sub in sub_list:
            cb_ld_client.delete_subscription(sub.id)
    except RequestException as e:
        logger.warning("Could not clean subscriptions completely")
        raise


def clear_context_broker(
    url: str = None,
    fiware_header: FiwareHeader = None,
    clear_registrations: bool = False,
    cb_client: ContextBrokerClient = None,
):
    """
    Function deletes all entities, registrations and subscriptions for a
    given fiware header. To use TLS connection you need to provide the cb_client parameter
    as an argument with the Session object including the certificate and private key.

    Note:
        Always clear the devices first because the IoT-Agent will otherwise
        through errors if it cannot find its registration anymore.

    Args:
        url: Url of the context broker service
        fiware_header: header of the tenant
        cb_client: enables TLS communication if created with Session object, only needed
                    for self-signed certificates
        clear_registrations: Determines whether registrations should be deleted.
                             If registrations are deleted while devices with commands
                             still exist, these devices become unreachable.
                             Only set to true once such devices are cleared.
    Returns:
        None
    """
    assert url or cb_client, "Either url or client object must be given"
    # create client
    if cb_client is None:
        client = ContextBrokerClient(url=url, fiware_header=fiware_header)
    else:
        client = cb_client

    # clear registrations
    if clear_registrations:
        for reg in client.get_registration_list():
            client.delete_registration(registration_id=reg.id)
        assert len(client.get_registration_list()) == 0

    # clean entities
    client.delete_entities(entities=client.get_entity_list())

    # clear subscriptions
    for sub in client.get_subscription_list():
        client.delete_subscription(subscription_id=sub.id)
    assert len(client.get_subscription_list()) == 0


def clear_iot_agent(
    url: Union[str, AnyHttpUrl] = None,
    fiware_header: FiwareHeader = None,
    iota_client: IoTAClient = None,
):
    """
    Function deletes all device groups and devices for a
    given fiware header. To use TLS connection you need to provide the iota_client parameter
    as an argument with the Session object including the certificate and private key.

    Args:
        url: Url of the iot agent service
        fiware_header: header of the tenant
        iota_client: enables TLS communication if created with Session object, only needed for self-signed certificates

    Returns:
        None
    """
    assert url or iota_client, "Either url or client object must be given"
    # create client
    if iota_client is None:
        client = IoTAClient(url=url, fiware_header=fiware_header)
    else:
        client = iota_client

    # clear registrations
    for device in client.get_device_list():
        client.delete_device(device_id=device.device_id)
    assert len(client.get_device_list()) == 0

    # clear groups
    for group in client.get_group_list():
        client.delete_group(resource=group.resource, apikey=group.apikey)
    assert len(client.get_group_list()) == 0


def clear_quantumleap(
    url: str = None,
    fiware_header: FiwareHeader = None,
    ql_client: QuantumLeapClient = None,
):
    """
    Function deletes all data for a given fiware header. To use TLS connection you need to provide the ql_client parameter
    as an argument with the Session object including the certificate and private key.
    Args:
        url: Url of the quantumleap service
        fiware_header: header of the tenant
        ql_client: enables TLS communication if created with Session object, only needed for self-signed certificates

    Returns:
        None
    """

    def handle_emtpy_db_exception(err: RequestException) -> None:
        """
        When the database is empty for request quantumleap returns a 404
        error with a error message. This will be handled here
        evaluating the empty database error as 'OK'

        Args:
            err: exception raised by delete function
        """
        if (
            err.response.status_code == 404
            and err.response.json().get("error", None) == "Not Found"
        ):
            pass
        else:
            raise

    assert url or ql_client, "Either url or client object must be given"
    # create client
    if ql_client is None:
        client = QuantumLeapClient(url=url, fiware_header=fiware_header)
    else:
        client = ql_client

    # clear data
    entities = []
    try:
        entities = client.get_entities()
    except RequestException as err:
        handle_emtpy_db_exception(err)

    # will be executed for all found entities
    for entity in entities:
        client.delete_entity(entity_id=entity.entityId, entity_type=entity.entityType)


def clear_all(
    *,
    fiware_header: FiwareHeader = None,
    cb_url: str = None,
    iota_url: Union[str, List[str]] = None,
    ql_url: str = None,
    cb_client: ContextBrokerClient = None,
    iota_client: IoTAClient = None,
    ql_client: QuantumLeapClient = None
):
    """
    Clears all services that a url is provided for.
    If cb_url is provided, the registration will also be deleted.

    Args:
        fiware_header:
        cb_url: url of the context broker service
        iota_url: url of the IoT-Agent service
        ql_url: url of the QuantumLeap service
        cb_client: enables TLS communication if created with Session object, only needed
         for self-signed certificates
        iota_client: enables TLS communication if created with Session object, only needed
         for self-signed certificates
        ql_client: enables TLS communication if created with Session object, only needed
         for self-signed certificates

    Returns:
        None
    """
    if iota_url is not None or iota_client is not None:
        if iota_url is None:
            # loop client
            if isinstance(iota_client, IoTAClient):
                iota_client = [iota_client]
            for client in iota_client:
                clear_iot_agent(fiware_header=fiware_header, iota_client=client)
        else:
            if isinstance(iota_url, (str, AnyUrl)):
                iota_url = [iota_url]
            for url in iota_url:
                clear_iot_agent(url=url, fiware_header=fiware_header)

    if cb_url is not None or cb_client is not None:
        clear_context_broker(
            url=cb_url,
            fiware_header=fiware_header,
            cb_client=cb_client,
            clear_registrations=True,
        )

    if ql_url is not None or ql_client is not None:
        clear_quantumleap(url=ql_url, fiware_header=fiware_header, ql_client=ql_client)


def clean_test(
    *,
    fiware_service: str,
    fiware_servicepath: str,
    cb_url: str = None,
    iota_url: Union[str, List[str]] = None,
    ql_url: str = None,
    cb_client: ContextBrokerClient = None,
    iota_client: IoTAClient = None,
    ql_client: QuantumLeapClient = None
) -> Callable:
    """
    Decorator to clean up the server before and after the test

    Note:
        This does not substitute a proper TearDown method, because a failing
        test will not execute the clean up after the error. Since this would
        mean an unnecessary error handling. We actually want a test to fail
        with proper messages.

    Args:
        fiware_service: tenant
        fiware_servicepath: tenant path
        cb_url: url of context broker service
        iota_url: url of IoT-Agent service
        ql_url: url of quantumleap service
        cb_client: enables TLS communication if created with Session object, only needed for self-signed certificates
        iota_client: enables TLS communication if created with Session object, only needed for self-signed certificates
        ql_client: enables TLS communication if created with Session object, only needed for self-signed certificates

    Returns:
        Decorator for clean tests
    """
    fiware_header = FiwareHeader(
        service=fiware_service, service_path=fiware_servicepath
    )
    clear_all(
        fiware_header=fiware_header,
        cb_url=cb_url,
        iota_url=iota_url,
        ql_url=ql_url,
        cb_client=cb_client,
        iota_client=iota_client,
        ql_client=ql_client,
    )

    # Inner decorator function
    def decorator(func):
        #  Wrapper function for the decorated function
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    clear_all(
        fiware_header=fiware_header,
        cb_url=cb_url,
        iota_url=iota_url,
        ql_url=ql_url,
        cb_client=cb_client,
        iota_client=iota_client,
        ql_client=ql_client,
    )

    return decorator



"""
Filter functions to keep client code clean and easy to use.
"""

from typing import List, Union
from filip.clients.ngsi_v2 import ContextBrokerClient
from filip.models import FiwareHeader
from filip.models.ngsi_v2.iot import Device, ServiceGroup
from filip.models.ngsi_v2.subscriptions import Subscription
from requests.exceptions import RequestException


def filter_device_list(
    devices: List[Device],
    device_ids: Union[str, List[str]] = None,
    entity_names: Union[str, List[str]] = None,
    entity_types: Union[str, List[str]] = None,
) -> List[Device]:
    """
    Filter the given device list based on conditions

    Args:
        devices: device list that need to be filtered
        device_ids: A list of device_id as filter condition
        entity_names: A list of entity_name (e.g. entity_id) as filter condition.
        entity_types: A list of entity_type as filter condition

    Returns:
        List of matching devices
    """
    if device_ids:
        if isinstance(device_ids, (list, str)):
            if isinstance(device_ids, str):
                device_ids = [device_ids]
            devices = [device for device in devices if device.device_id in device_ids]
        else:
            raise TypeError("device_ids must be a string or a list of strings!")

    if entity_names:
        if isinstance(entity_names, (list, str)):
            if isinstance(entity_names, str):
                entity_names = [entity_names]
            devices = [
                device for device in devices if device.entity_name in entity_names
            ]
        else:
            raise TypeError("entity_names must be a string or a list of strings!")

    if entity_types:
        if isinstance(entity_types, (list, str)):
            if isinstance(entity_types, str):
                entity_types = [entity_types]
            devices = [
                device for device in devices if device.entity_type in entity_types
            ]
        else:
            raise TypeError("entity_types must be a string or a list of strings!")

    return devices


def filter_subscriptions_by_entity(
    entity_id: str,
    entity_type: str,
    url: str = None,
    fiware_header: FiwareHeader = None,
    subscriptions: List[Subscription] = None,
) -> List[Subscription]:
    """
    Function that filters subscriptions based on the entity id or id pattern
    and entity type or type pattern. The function can be used in two ways,
    wither pass list of subscriptions to filter based on entity or directly pass
    client information to filter subscriptions in a single request.

    Args:
        entity_id: Id of the entity to be matched
        entity_type: Type of the entity to be matched
        url: Url of the context broker service
        fiware_header: Fiware header of the tenant
        subscriptions: List of subscriptions to filter
    Returns:
        list of subscriptions by entity
    """
    if not subscriptions:
        client = ContextBrokerClient(url=url, fiware_header=fiware_header)
        subscriptions = client.get_subscription_list()
    filtered_subscriptions = []
    for subscription in subscriptions:
        for entity in subscription.subject.entities:
            if entity.id == entity_id or (
                entity.idPattern is not None and entity.idPattern.match(entity_id)
            ):
                if entity.type == entity_type or (
                    entity.typePattern is not None
                    and entity.typePattern.match(entity_type)
                ):
                    filtered_subscriptions.append(subscription)
    return filtered_subscriptions


def filter_group_list(
    group_list: List[ServiceGroup],
    resources: Union[str, List[str]] = None,
    apikeys: Union[str, List[str]] = None,
) -> List[ServiceGroup]:
    """
    Filter service group based on resource and apikey.

    Args:
        group_list: The list of service groups that need to be filtered
        resources: see ServiceGroup model
        apikeys: see ServiceGroup

    Returns: a single service group or Not Found Error
    """
    if resources:
        if isinstance(resources, (list, str)):
            if isinstance(resources, str):
                resources = [resources]
            group_list = [group for group in group_list if group.resource in resources]
        else:
            raise TypeError("resources must be a string or a list of strings!")

    if apikeys:
        if isinstance(apikeys, (list, str)):
            if isinstance(apikeys, str):
                apikeys = [apikeys]
            group_list = [group for group in group_list if group.apikey in apikeys]
        else:
            raise TypeError("apikeys must be a string or a list of strings!")

    return group_list



"""
Data models for interacting with FIWARE's time series-api (aka QuantumLeap)
"""

from __future__ import annotations
import logging
from typing import Any, List, Union
from datetime import datetime
import numpy as np
import pandas as pd
from aenum import Enum
from pydantic import ConfigDict, BaseModel, Field


logger = logging.getLogger(__name__)


class TimeSeriesBase(BaseModel):
    """
    Base model for other time series api models
    """

    index: Union[List[datetime], datetime] = Field(
        default=None,
        description="Array of the timestamps which are indexes of the response "
        "for the requested data. It's a parallel array to 'values'."
        " The timestamp will be in the ISO8601 format "
        "(e.g. 2010-10-10T07:09:00.792) or in milliseconds since "
        "epoch whichever format was used in the input "
        "(notification), but ALWAYS in UTC. When using aggregation "
        "options, the format of this remains the same, only the "
        "semantics will change. For example, if aggrPeriod is day, "
        "each index will be a valid timestamp of a moment in the "
        "corresponding day.",
    )


class TimeSeriesHeader(TimeSeriesBase):
    """
    Model to describe an available entity in the time series api
    """

    model_config = ConfigDict(populate_by_name=True)
    # aliases are required due to formally inconsistencies in the api-specs
    entityId: str = Field(
        default=None,
        alias="id",
        description="The entity id the time series api."
        "If the id is unique among all entity "
        "types, this could be used to uniquely "
        "identify the entity instance. Otherwise,"
        " you will have to use the entityType "
        "attribute to resolve ambiguity.",
    )
    entityType: str = Field(
        default=None, alias="type", description="The type of an entity"
    )


class IndexedValues(BaseModel):
    """
    Model for time indexed values
    """

    values: List[Any] = Field(
        default=None,
        description="Array of values of the selected attribute, in the same "
        "corresponding order of the 'index' array. When using "
        "aggregation options, the format of this remains the same, "
        "only the semantics will change. For example, if "
        "aggrPeriod is day, each value of course may not "
        "correspond to original measurements but rather the "
        "aggregate of measurements in each day.",
    )


class AttributeValues(IndexedValues):
    """
    Model for indexed values that contain attribute name
    """

    attrName: str = Field(title="Attribute name", description="")


class TimeSeries(TimeSeriesHeader):
    """
    Model for time series data
    """

    model_config = ConfigDict(populate_by_name=True)
    attributes: List[AttributeValues] = None

    def extend(self, other: TimeSeries) -> None:
        """
        Extends the current `TimeSeries` object with an other
        `TimeSeries` object. With the same format.

        Args:
            other: TimeSeries Object that will be added to the original object

        Returns:
            None

        Raises:
            Assertion Error: if header fields do not fit or if index is not
                rising
        """
        assert self.entityId == other.entityId
        assert self.entityType == other.entityType
        assert self.index[-1] < other.index[0]

        for attr, other_attr in zip(self.attributes, other.attributes):
            assert attr.attrName == other_attr.attrName
            attr.values.extend(other_attr.values)
        self.index.extend(other.index)

    def to_pandas(self) -> pd.DataFrame:
        """
        Converts time series data to pandas dataframe
        Returns:
            pandas.DataFrame
        """
        index = pd.Index(data=self.index, name="datetime")
        attr_names = [attr.attrName for attr in self.attributes]
        values = np.array([attr.values for attr in self.attributes]).transpose()
        columns = pd.MultiIndex.from_product(
            [[self.entityId], [self.entityType], attr_names],
            names=["entityId", "entityType", "attribute"],
        )

        return pd.DataFrame(data=values, index=index, columns=columns)


class AggrMethod(str, Enum):
    """
    Aggregation Methods
    """

    _init_ = "value __doc__"
    COUNT = "count", "Number of Entries"
    SUM = "sum", "Sum"
    AVG = "avg", "Average"
    MIN = "min", "Minimum"
    MAX = "max", "Maximum"


class AggrPeriod(str, Enum):
    """
    Aggregation Periods
    """

    _init_ = "value __doc__"
    YEAR = "year", "year"
    MONTH = "month", "month"
    DAY = "day", "day"
    HOUR = "hour", "hour"
    MINUTE = "minute", "minute"
    SECOND = "second", "second"


class AggrScope(str, Enum):
    """
    Aggregation Periods
    When the query results cover historical data for
    multiple entities instances, you can define the aggregation method to be
    applied for each entity instance [entity] or across them [global].
    """

    _init_ = "value __doc__"
    ENTITY = "entity", "Entity (default)"
    GLOBAL = "global", "Global"



"""
Module for client specific exceptions
"""
import requests.models
from requests import RequestException


class BaseHttpClientException(RequestException):
    """
    Base exception class for all HTTP clients. The response of a request will be available in the exception.

    Args:
        message (str): Error message
        response (Response): Response object
    """
    def __init__(self, message: str, response: requests.models.Response):
        super().__init__(message)
        self.response = response



"""
IoT-Agent Module for API Client
"""

from __future__ import annotations

import json
from copy import deepcopy
from typing import List, Dict, Set, TYPE_CHECKING, Union, Optional
import warnings
from urllib.parse import urljoin
import requests
from pydantic import AnyHttpUrl
from pydantic.type_adapter import TypeAdapter
from filip.config import settings
from filip.clients.base_http_client import BaseHttpClient
from filip.clients.exceptions import BaseHttpClientException
from filip.models.base import FiwareHeader
from filip.models.ngsi_v2.iot import Device, ServiceGroup

from filip.utils.filter import filter_device_list, filter_group_list

if TYPE_CHECKING:
    from filip.clients.ngsi_v2.cb import ContextBrokerClient


class IoTAClient(BaseHttpClient):
    """
    Client for FIWARE IoT-Agents. The implementation follows the API
    specifications from here:
    https://iotagent-node-lib.readthedocs.io/en/latest/

    Args:
        url: Url of IoT-Agent
        session (requests.Session):
        fiware_header (FiwareHeader): fiware service and fiware service path
        **kwargs (Optional): Optional arguments that ``request`` takes.
    """

    def __init__(
        self,
        url: str = None,
        *,
        session: requests.Session = None,
        fiware_header: FiwareHeader = None,
        **kwargs,
    ):
        # set service url
        url = url or settings.IOTA_URL
        super().__init__(
            url=url, session=session, fiware_header=fiware_header, **kwargs
        )

    # ABOUT API
    def get_version(self) -> Dict:
        """
        Gets version of IoT Agent

        Returns:
            Dictionary with response
        """
        url = urljoin(self.base_url, "iot/about")
        try:
            res = self.get(url=url, headers=self.headers)
            if res.ok:
                return res.json()
            res.raise_for_status()
        except requests.RequestException as err:
            self.logger.error(err)
            msg = "Could not retrieve version because of following reason: " + str(
                err.args[0]
            )
            raise BaseHttpClientException(message=msg, response=err.response) from err

    # SERVICE GROUP API
    def post_groups(
        self,
        service_groups: Union[ServiceGroup, List[ServiceGroup]],
        update: bool = False,
    ):
        """
        Creates a set of service groups for the given service and service_path.
        The service_group and subservice information will taken from the
        headers, overwriting any preexisting values.

        Args:
            service_groups (list of ServiceGroup): Service groups that will be
            posted to the agent's API
            update (bool): If service group already exists try to update its

        Returns:
            None
        """
        if not isinstance(service_groups, list):
            service_groups = [service_groups]
        for group in service_groups:
            if group.service:
                assert (
                    group.service == self.headers["fiware-service"]
                ), "Service group service does not math fiware service"
            if group.subservice:
                assert (
                    group.subservice == self.headers["fiware-servicepath"]
                ), "Service group subservice does not match fiware service path"

        url = urljoin(self.base_url, "iot/services")
        headers = self.headers
        data = {
            "services": [
                group.model_dump(exclude={"service", "subservice"}, exclude_none=True)
                for group in service_groups
            ]
        }
        try:
            res = self.post(url=url, headers=headers, json=data)
            if res.ok:
                self.logger.info("Services successfully posted")
            elif res.status_code == 409:
                self.logger.warning(res.text)
                if len(service_groups) > 1:
                    self.logger.info(
                        "Trying to split bulk operation into " "single operations"
                    )
                    for group in service_groups:
                        self.post_group(service_group=group, update=update)
                elif update is True:
                    self.update_group(service_group=service_groups[0], fields=None)
                else:
                    res.raise_for_status()
            else:
                res.raise_for_status()
        except requests.RequestException as err:
            self.logger.error(err)
            msg = "Could not post group because of following reason: " + str(
                err.args[0]
            )
            raise BaseHttpClientException(message=msg, response=err.response) from err

    def post_group(self, service_group: ServiceGroup, update: bool = False):
        """
        Single service registration but using the bulk operation in background

        Args:
            service_group (ServiceGroup): Service that will be posted to the
            agent's API
            update (bool):

        Returns:
            None
        """
        return self.post_groups(service_groups=[service_group], update=update)

    def get_group_list(self) -> List[ServiceGroup]:
        r"""
        Retrieves service_group groups from the database. If the servicepath
        header has the wildcard expression, /\*, all the subservices for the
        service_group are returned. The specific subservice parameters are
        returned in any other case.

        Returns:

        """
        url = urljoin(self.base_url, "iot/services")
        headers = self.headers
        try:
            res = self.get(url=url, headers=headers)
            if res.ok:
                ta = TypeAdapter(List[ServiceGroup])
                return ta.validate_python(res.json()["services"])
            res.raise_for_status()
        except requests.RequestException as err:
            self.logger.error(err)
            msg = "Could not retrieve group list because of following reason: " + str(
                err.args[0]
            )
            raise BaseHttpClientException(message=msg, response=err.response) from err

    def get_group(self, *, resource: str, apikey: str) -> ServiceGroup:
        """
        Retrieves service_group groups from the database based on resource and
        apikey
        Args:
            resource:
            apikey:
        Returns:

        """
        groups = self.get_group_list()
        groups = filter_group_list(
            group_list=groups, resources=resource, apikeys=apikey
        )
        if len(groups) == 1:
            group = groups[0]
            return group
        elif len(groups) == 0:
            raise KeyError(
                f"Service group with resource={resource} and apikey={apikey} was not found"
            )
        else:
            raise NotImplementedError(
                "There is a wierd error, try get_group_list() for debugging"
            )

    def update_groups(
        self,
        *,
        service_groups: Union[ServiceGroup, List[ServiceGroup]],
        add: False,
        fields: Union[Set[str], List[str]] = None,
    ) -> None:
        """
        Bulk operation for service group update.
        Args:
            fields:
            service_groups:
            add:

        Returns:

        """
        if not isinstance(service_groups, list):
            service_groups = [service_groups]
        for group in service_groups:
            self.update_group(service_group=group, fields=fields, add=add)

    def update_group(
        self,
        *,
        service_group: ServiceGroup,
        fields: Union[Set[str], List[str]] = None,
        add: bool = True,
    ):
        """
        Modifies the information for a service group configuration, identified
        by the resource and apikey query parameters. Takes a service group body
        as the payload. The body does not have to be complete: for incomplete
        bodies, just the existing attributes will be updated

        Args:
            service_group (ServiceGroup): Service to update.
            fields: Fields of the service_group to update. If 'None' all allowed
            fields will be updated
            add:
        Returns:
            None
        """
        if fields:
            if isinstance(fields, list):
                fields = set(fields)
        else:
            fields = None
        url = urljoin(self.base_url, "iot/services")
        headers = self.headers
        params = service_group.model_dump(include={"resource", "apikey"})
        try:
            res = self.put(
                url=url,
                headers=headers,
                params=params,
                json=service_group.model_dump(
                    include=fields, exclude={"service", "subservice"}, exclude_none=True
                ),
            )
            if res.ok:
                self.logger.info("ServiceGroup updated!")
            elif (res.status_code == 404) & (add is True):
                self.post_group(service_group=service_group)
            else:
                res.raise_for_status()
        except requests.RequestException as err:
            self.logger.error(err)
            msg = "Could not update group because of following reason: " + str(
                err.args[0]
            )
            raise BaseHttpClientException(message=msg, response=err.response) from err

    def delete_group(self, *, resource: str, apikey: str):
        """
        Deletes a service group in in the IoT-Agent

        Args:
            resource:
            apikey:

        Returns:

        """
        url = urljoin(self.base_url, "iot/services")
        headers = self.headers
        params = {"resource": resource, "apikey": apikey}
        try:
            res = self.delete(url=url, headers=headers, params=params)
            if res.ok:
                self.logger.info(
                    "ServiceGroup with resource: '%s' and "
                    "apikey: '%s' successfully deleted!",
                    resource,
                    apikey,
                )
            else:
                res.raise_for_status()
        except requests.RequestException as err:
            self.logger.error(err)
            msg = (
                f"Could not delete ServiceGroup with resource "
                f"'{resource}' and apikey '{apikey}' because of following reason: {str(err.args[0])}"
            )
            raise BaseHttpClientException(message=msg, response=err.response) from err

    # DEVICE API
    def post_devices(
        self, *, devices: Union[Device, List[Device]], update: bool = False
    ) -> None:
        """
        Post a device from the device registry. No payload is required
        or received.
        If a device already exists in can be updated with update = True
        Args:
            devices (list of Devices):
            update (bool):  Whether if the device is already existent it
            should be updated
        Returns:
            None
        """
        if not isinstance(devices, list):
            devices = [devices]
        url = urljoin(self.base_url, "iot/devices")
        headers = self.headers

        data = {
            "devices": [
                json.loads(device.model_dump_json(exclude_none=True))
                for device in devices
            ]
        }
        try:
            res = self.post(url=url, headers=headers, json=data)
            if res.ok:
                self.logger.info("Devices successfully posted!")
            else:
                res.raise_for_status()
        except requests.RequestException as err:
            if update:
                return self.update_devices(devices=devices, add=False)
            self.logger.error(err)
            msg = "Could not post devices because of following reason: " + str(
                err.args[0]
            )
            raise BaseHttpClientException(message=msg, response=err.response) from err

    def post_device(self, *, device: Device, update: bool = False) -> None:
        """
        Post a device configuration to the IoT-Agent

        Args:
            device: IoT device configuration to send
            update: update device if configuration already exists

        Returns:
            None
        """
        return self.post_devices(devices=[device], update=update)

    def get_device_list(
        self,
        *,
        limit: int = None,
        offset: int = None,
        device_ids: Union[str, List[str]] = None,
        entity_names: Union[str, List[str]] = None,
        entity_types: Union[str, List[str]] = None,
    ) -> List[Device]:
        """
        Returns a list of all the devices in the device registry with all
        its data. The IoTAgent now only supports "limit" and "offset" as
        request parameters.

        Args:
            limit:
                if present, limits the number of devices returned in the
                list. Must be a number between 1 and 1000.
            offset:
                if present, skip that number of devices from the original
                query.
            device_ids:
                List of device_ids. If given, only devices with matching ids
                will be returned
            entity_names:
                The entity_ids of the devices. If given, only the devices
                with the specified entity_id will be returned
            entity_types:
                The entity_type of the device. If given, only the devices
                with the specified entity_type will be returned

        Returns:
            List of matching devices
        """
        params = {}
        if limit:
            if not 1 < limit < 1000:
                self.logger.error("'limit' must be an integer between 1 and " "1000!")
                raise ValueError
            else:
                params["limit"] = limit
        if offset:
            if not isinstance(offset, int):
                self.logger.error("'offset' must be an integer!")
                raise ValueError
            else:
                params["offset"] = offset
        url = urljoin(self.base_url, "iot/devices")
        headers = self.headers
        try:
            res = self.get(url=url, headers=headers, params=params)
            if res.ok:
                ta = TypeAdapter(List[Device])
                devices = ta.validate_python(res.json()["devices"])
                # filter by device_ids, entity_names or entity_types
                devices = filter_device_list(
                    devices, device_ids, entity_names, entity_types
                )
                return devices
            res.raise_for_status()
        except requests.RequestException as err:
            self.logger.error(err)
            msg = (
                "Not able to retrieve the device list because of the following reason:"
                + str(err.args[0])
            )
            raise BaseHttpClientException(message=msg, response=err.response) from err

    def get_device(self, *, device_id: str) -> Device:
        """
        Returns all the information about a particular device.

        Args:
            device_id:
        Raises:
            requests.RequestException, if device does not exist
        Returns:
            Device

        """
        url = urljoin(self.base_url, f"iot/devices/{device_id}")
        headers = self.headers
        try:
            res = self.get(url=url, headers=headers)
            if res.ok:
                return Device.model_validate(res.json())
            res.raise_for_status()
        except requests.RequestException as err:
            self.logger.error(err)

            msg = f"Device '{device_id}' was not found because of the following reason: {str(err.args[0])}"
            raise BaseHttpClientException(message=msg, response=err.response) from err

    def update_device(self, *, device: Device, add: bool = True) -> None:
        """
        Updates a device from the device registry.
        Adds, removes attributes from the device entry and changes
        attributes values.
        It does not change device settings (endpoint,..) and only adds
        attributes to the corresponding entity, their it does not
        change any attribute value and does not delete removed attributes

        Args:
            device:
            add (bool): If device not found add it
        Returns:
            None
        """
        url = urljoin(self.base_url, f"iot/devices/{device.device_id}")
        headers = self.headers
        try:
            res = self.put(
                url=url,
                headers=headers,
                json=device.model_dump(
                    include={"attributes", "lazy", "commands", "static_attributes"},
                    exclude_none=True,
                ),
            )
            if res.ok:
                self.logger.info("Device '%s' successfully updated!", device.device_id)
            elif (res.status_code == 404) & (add is True):
                self.post_device(device=device, update=False)
            else:
                res.raise_for_status()
        except requests.RequestException as err:
            self.logger.error(err)
            msg = f"Could not update device '{device.device_id}' because of the following reason: {str(err.args[0])} "
            raise BaseHttpClientException(message=msg, response=err.response) from err

    def update_devices(
        self, *, devices: Union[Device, List[Device]], add: False
    ) -> None:
        """
        Bulk operation for device update.
        Args:
            devices:
            add:

        Returns:

        """
        if not isinstance(devices, list):
            devices = [devices]
        for device in devices:
            self.update_device(device=device, add=add)

    def delete_device(
        self,
        *,
        device_id: str,
        cb_url: AnyHttpUrl = settings.CB_URL,
        delete_entity: bool = False,
        force_entity_deletion: bool = False,
        cb_client: ContextBrokerClient = None,
    ) -> None:
        """
        Remove a device from the device registry. No payload is required
        or received.

        Args:
            device_id: str, ID of Device
            delete_entity:  False -> Only delete the device entry,
                                     the automatically created and linked
                                     context-entity will continue to
                                     exist in Fiware
                            True -> Also delete the automatically
                                    created and linked context-entity
                                    If multiple devices are linked to this
                                    entity, this operation is not executed and
                                    an exception is raised
            force_entity_deletion:
                bool, if delete_entity is true and multiple devices are linked
                to the linked entity, delete it and do not raise an error
            cb_client (ContextBrokerClient):
                Corresponding ContextBrokerClient object for entity manipulation
            cb_url (AnyHttpUrl):
                Url of the ContextBroker where the entity is found.
                This will autogenerate an CB-Client, mirroring the information
                of the IoTA-Client, e.g. FiwareHeader, and other headers
                (not recommended!)

        Returns:
            None
        """
        url = urljoin(
            self.base_url,
            f"iot/devices/{device_id}",
        )
        headers = self.headers

        device = self.get_device(device_id=device_id)

        try:
            res = self.delete(url=url, headers=headers)
            if res.ok:
                self.logger.info("Device '%s' successfully deleted!", device_id)
            else:
                res.raise_for_status()
        except requests.RequestException as err:
            msg = f"Could not delete device {device_id}!"
            raise BaseHttpClientException(message=msg, response=err.response) from err

        if delete_entity:
            # An entity can technically belong to multiple devices
            # Only delete the entity if
            devices = self.get_device_list(entity_names=[device.entity_name])

            # Zero because we count the remaining devices
            if len(devices) > 0 and not force_entity_deletion:
                raise Exception(
                    f"The corresponding entity to the device "
                    f"{device_id} was not deleted because it is "
                    f"linked to multiple devices. "
                )
            else:
                cb_client_local = None
                try:
                    from filip.clients.ngsi_v2 import ContextBrokerClient

                    if cb_client:
                        cb_client_local = deepcopy(cb_client)
                    else:
                        warnings.warn(
                            "No `ContextBrokerClient` "
                            "object provided! Will try to generate "
                            "one. This usage is not recommended."
                        )

                        cb_client_local = ContextBrokerClient(
                            url=cb_url,
                            fiware_header=self.fiware_headers,
                            headers=headers,
                        )

                    cb_client_local.delete_entity(
                        entity_id=device.entity_name, entity_type=device.entity_type
                    )

                except requests.RequestException as err:
                    # Do not throw an error
                    # It is only important that the entity does not exist after
                    # this methode, not if this methode actively deleted it
                    pass

                if cb_client_local:
                    cb_client_local.close()

    def patch_device(
        self,
        device: Device,
        patch_entity: bool = True,
        cb_client: ContextBrokerClient = None,
        cb_url: AnyHttpUrl = settings.CB_URL,
    ) -> None:
        """
        Updates a device state in Fiware, if the device does not exist it
        is created, else its values are updated.
        If the device settings were changed the device and
        entity are deleted and re-added.

        If patch_entity is true the corresponding entity in the ContextBroker is
        also correctly updated. Else only new attributes are added there.

        Args:
            device (Device): Device to be posted to /updated in Fiware
            patch_entity (bool): If true the corresponding entity is
                completely synced
            cb_client (ContextBrokerClient):
                Corresponding ContextBrokerClient object for entity manipulation
            cb_url (AnyHttpUrl):
                Url of the ContextBroker where the entity is found.
                This will autogenerate an CB-Client, mirroring the information
                of the IoTA-Client, e.g. FiwareHeader, and other headers
                (not recommended!)

        Returns:
            None
        """
        try:
            live_device = self.get_device(device_id=device.device_id)
        except requests.RequestException:
            # device does not exist yet, post it
            self.post_device(device=device)
            return

        # if the device settings were changed we need to delete the device
        # and repost it
        settings_dict = {
            "device_id",
            "service",
            "service_path",
            "entity_name",
            "entity_type",
            "timestamp",
            "apikey",
            "endpoint",
            "protocol",
            "transport",
            "expressionLanguage",
        }

        live_settings = live_device.model_dump(include=settings_dict)
        new_settings = device.model_dump(include=settings_dict)

        if not live_settings == new_settings:
            self.delete_device(
                device_id=device.device_id,
                delete_entity=True,
                force_entity_deletion=True,
                cb_client=cb_client,
            )
            self.post_device(device=device)
            return

        # We are at a state where the device exists, but only attributes were
        # changed.
        # we need to update the device, and the context entry separately,
        # as update device only takes over a part of the changes to the
        # ContextBroker.

        # update device
        self.update_device(device=device)

        # update context entry
        # 1. build context entity from information in device
        # 2. patch it
        from filip.models.ngsi_v2.context import ContextEntity, NamedContextAttribute

        def build_context_entity_from_device(device: Device) -> ContextEntity:
            from filip.models.base import DataType

            entity = ContextEntity(id=device.entity_name, type=device.entity_type)

            for command in device.commands:
                entity.add_attributes(
                    [
                        # Command attribute will be registered by the device_update
                        NamedContextAttribute(
                            name=f"{command.name}_info", type=DataType.COMMAND_RESULT
                        ),
                        NamedContextAttribute(
                            name=f"{command.name}_status", type=DataType.COMMAND_STATUS
                        ),
                    ]
                )
            for attribute in device.attributes:
                entity.add_attributes(
                    [
                        NamedContextAttribute(
                            name=attribute.name,
                            type=DataType.STRUCTUREDVALUE,
                            metadata=attribute.metadata,
                        )
                    ]
                )
            for static_attribute in device.static_attributes:
                entity.add_attributes(
                    [
                        NamedContextAttribute(
                            name=static_attribute.name,
                            type=static_attribute.type,
                            value=static_attribute.value,
                            metadata=static_attribute.metadata,
                        )
                    ]
                )
            return entity

        if patch_entity:
            from filip.clients.ngsi_v2 import ContextBrokerClient

            if cb_client:
                cb_client_local = deepcopy(cb_client)
            else:
                warnings.warn(
                    "No `ContextBrokerClient` object provided! "
                    "Will try to generate one. "
                    "This usage is not recommended."
                )

                cb_client_local = ContextBrokerClient(
                    url=cb_url, fiware_header=self.fiware_headers, headers=self.headers
                )

            cb_client_local.override_entity(
                entity=build_context_entity_from_device(device)
            )
            cb_client_local.close()

    def does_device_exists(self, device_id: str) -> bool:
        """
        Test if a device with the given id exists in Fiware
        Args:
            device_id (str)
        Returns:
            bool
        """
        try:
            self.get_device(device_id=device_id)
            return True
        except requests.RequestException as err:
            if err.response is None or not err.response.status_code == 404:
                self.logger.error(err)
                msg = f"Could not check device status because of the following reason: {str(err.args[0])}"
                raise BaseHttpClientException(
                    message=msg, response=err.response
                ) from err
            return False

    # LOG API
    def get_loglevel_of_agent(self):
        """
        Get current loglevel of agent
        Returns:

        """
        url = urljoin(self.base_url, "admin/log")
        headers = self.headers.copy()
        del headers["fiware-service"]
        del headers["fiware-servicepath"]
        try:
            res = self.get(url=url, headers=headers)
            if res.ok:
                return res.json()["level"]
            res.raise_for_status()
        except requests.RequestException as err:
            self.log_error(err=err)
            msg = f"Could not check for loglevel status because of the following reason: {str(err.args[0])}"
            raise BaseHttpClientException(message=msg, response=err.response) from err

    def change_loglevel_of_agent(self, level: str):
        """
        Change current loglevel of agent

        Args:
            level:

        Returns:

        """
        level = level.upper()
        if level not in ["INFO", "ERROR", "FATAL", "DEBUG", "WARNING"]:
            raise KeyError("Given log level is not supported")

        url = urljoin(self.base_url, "admin/log")
        headers = self.headers.copy()
        new_loglevel = {"level": level}
        del headers["fiware-service"]
        del headers["fiware-servicepath"]
        try:
            res = self.put(url=url, headers=headers, params=new_loglevel)
            if res.ok:
                self.logger.info(
                    "Loglevel of agent at %s " "changed to '%s'",
                    self.base_url,
                    new_loglevel,
                )
            else:
                res.raise_for_status()
        except requests.RequestException as err:
            self.log_error(err=err)
            msg = f"Could not change loglevel because of the following reason: {str(err.args[0])}"
            raise BaseHttpClientException(message=msg, response=err.response) from err



"""Module containing the models describing a semantic state"""

import uuid

import pydantic as pyd
import requests
from aenum import Enum
from typing import (
    List,
    Tuple,
    Dict,
    Type,
    TYPE_CHECKING,
    Optional,
    Union,
    Set,
    Iterator,
    Any,
)

import filip.models.ngsi_v2.iot as iot

# from filip.models.ngsi_v2.iot import ExpressionLanguage, TransportProtocol
from filip.models.base import DataType, NgsiVersion
from filip.utils.validators import FiwareRegex
from filip.models.ngsi_v2.context import (
    ContextEntity,
    NamedContextAttribute,
    NamedCommand,
)

from filip.models import FiwareHeader
from pydantic import ConfigDict, BaseModel, Field
from filip.config import settings
from filip.semantics.vocabulary.entities import DatatypeFields, DatatypeType
from filip.semantics.vocabulary_configurator import (
    label_blacklist,
    label_char_whitelist,
)

if TYPE_CHECKING:
    from filip.semantics.semantics_manager import SemanticsManager


class InstanceHeader(FiwareHeader):
    """
    Header of a SemanticClass instance, describes the Fiware Location were
    the instance will be / is saved.
    The header is not bound to one Fiware Setup, but can describe the
    exact location in the web
    """

    model_config = ConfigDict(frozen=True, use_enum_values=True)
    cb_url: str = Field(
        default=settings.CB_URL,
        description="Url of the ContextBroker from the Fiware " "setup",
    )
    iota_url: str = Field(
        default=settings.IOTA_URL,
        description="Url of the IoTABroker from the Fiware " "setup",
    )

    ngsi_version: NgsiVersion = Field(
        default=NgsiVersion.v2, description="Used Version in the " "Fiware setup"
    )

    def get_fiware_header(self) -> FiwareHeader:
        """
        Get a Filip FiwareHeader from the InstanceHeader
        """
        return FiwareHeader(service=self.service, service_path=self.service_path)


class InstanceIdentifier(BaseModel):
    """
    Each Instance of a SemanticClass posses a unique identifier that is
    directly linked to one Fiware entry
    """

    model_config = ConfigDict(frozen=True)
    id: str = Field(description="Id of the entry in Fiware")
    type: str = Field(description="Type of the entry in Fiware, equal to " "class_name")
    header: InstanceHeader = Field(
        description="describes the Fiware "
        "Location were the instance "
        "will be / is saved."
    )


class Datatype(DatatypeFields):
    """
    Model of a vocabulary/ontology Datatype used to validate assignments in
    DataFields
    """

    def value_is_valid(self, value: str) -> bool:
        """
        Test if value is valid for this datatype.
        Numbers are also given as strings

        Args:
            value (str): value to be tested

        Returns:
            bool
        """
        if self.type == "string":
            if len(self.allowed_chars) > 0:
                for char in value:
                    if char not in self.allowed_chars:
                        return False
            for char in self.forbidden_chars:
                if char in value:
                    return False
            return True

        if self.type == "number":
            if self.number_decimal_allowed:
                try:
                    number = float(value)
                except ValueError:
                    return False
            else:
                try:
                    number = int(value)
                except ValueError:
                    return False

            if not self.number_range_min == "/":
                if number < self.number_range_min:
                    return False
            if not self.number_range_max == "/":
                if number > self.number_range_max:
                    return False

            return True

        if self.type == "enum":
            return value in self.enum_values

        if self.type == "date":
            try:
                from dateutil.parser import parse

                parse(value, fuzzy=False)
                return True

            except ValueError:
                return False

        return True


class DevicePropertyInstanceLink(BaseModel):
    """
    SubProperties of a DeviceProperty, containing the information to which
    instance the DeviceProperty belongs.

    Modeled as a standalone model, to bypass the read-only logic of
    DeviceProperty
    """

    instance_identifier: Optional[InstanceIdentifier] = Field(
        default=None, description="Identifier of the instance holding this Property"
    )
    semantic_manager: Optional["SemanticsManager"] = Field(
        default=None, description="Link to the governing semantic_manager"
    )
    field_name: Optional[str] = Field(
        default=None,
        description="Name of the field to which this property was added "
        "in the instance",
    )


class DeviceProperty(BaseModel):
    """
    Model describing one specific property of an IoT device.
    It is either a command that can be executed or an attribute that can be read

    A property can only belong to one field of one instance. Assigning it to
    multiple fields will result in an error.
    """

    model_config = ConfigDict()

    name: str = Field("Internally used name in the IoT Device")
    _instance_link: DevicePropertyInstanceLink = DevicePropertyInstanceLink()
    """Additional properties describing the instance and field where this \
    property was added"""

    def _get_instance(self) -> "SemanticClass":
        """Get the instance object to which this property was added"""

        return self._instance_link.semantic_manager.get_instance(
            self._instance_link.instance_identifier
        )

    def _get_field_from_fiware(
        self, field_name: str, required_type: str
    ) -> NamedContextAttribute:
        """
        Retrieves live information about a field from the assigned instance
        from Fiware

        Args:
            field_name (str): Name of the to retrieving field
            required_type (str): Type that the retrieved field is required to
                                 have
        Raises:
            Exception;  if the instance or the field is not present in Fiware
                        (the instance  state was not yet saved)
            Exception;  The field_type does not match
        """

        if self._instance_link.field_name is None:
            raise Exception(
                "This DeviceProperty needs to be added to a "
                "device field of an SemanticDeviceClass instance "
                "and the state saved before this methode can be "
                "executed"
            )

        try:
            entity = self._instance_link.semantic_manager.get_entity_from_fiware(
                instance_identifier=self._instance_link.instance_identifier
            )
        except requests.RequestException:
            raise Exception(
                "The instance to which this property belongs is "
                "not yet present in Fiware, you need to save the "
                "state first"
            )
        try:
            attr = entity.get_attribute(field_name)
        except requests.RequestException:
            raise Exception(
                "This property was not yet saved in Fiware. "
                "You need to save the state first before this "
                "methode can be executed"
            )

        if not attr.type == required_type:
            raise Exception(
                "The field in Fiware has a wrong type, "
                "an uncaught naming conflict happened"
            )
        return attr

    def get_all_field_names(self, field_name: Optional[str] = None) -> List[str]:
        """
        Get all field names which this property creates in the fiware
        instance

        Args:
            field_name (Optional[str]): Name of the field to which the attribute
                is/will be added. If none is provided, the linked field name
                is used
        """
        pass


class Command(DeviceProperty):
    """
    Model describing a command property of an IoT device.

    The command will add three fields to the fiware instance:
        - name - Used to execute the command, function: send()
        - name_info - Used to retrieve the command result: get_info()
        - name_status - Used to see the current status: get_status()

    A command can only belong to one field of one instance. Assigning it to
    multiple fields will result in an error.
    """

    model_config = ConfigDict(frozen=True)

    def send(self):
        """
        Execute the command on the IoT device

        Raises:
            Exception: If the command was not yet saved to Fiware
        """
        client = self._instance_link.semantic_manager.get_client(
            self._instance_link.instance_identifier.header
        )

        context_command = NamedCommand(name=self.name, value="")
        identifier = self._instance_link.instance_identifier
        client.post_command(
            entity_id=identifier.id,
            entity_type=identifier.type,
            command=context_command,
        )
        client.close()

    def get_info(self) -> str:
        """
        Retrieve the executed command result from the IoT-Device

        Raises:
            Exception: If the command was not yet saved to Fiware
        """
        return self._get_field_from_fiware(
            field_name=f"{self.name}_info", required_type="commandResult"
        ).value

    def get_status(self):
        """
        Retrieve the executed command status from the IoT-Device

        Raises:
            Exception: If the command was not yet saved to Fiware
        """
        return self._get_field_from_fiware(
            field_name=f"{self.name}_status", required_type="commandStatus"
        ).value

    def get_all_field_names(self, field_name: Optional[str] = None) -> List[str]:
        """
        Get all the field names that this command will add to Fiware

        Args:
            field_name (Optional[str]): Not used, but needed in the signature
        """
        return [self.name, f"{self.name}_info", f"{self.name}_result"]


class DeviceAttributeType(str, Enum):
    """
    Retrieval type of the DeviceAttribute value from the IoT Device into Fiware
    """

    _init_ = "value __doc__"

    lazy = "lazy", "The value is only read out if it is requested"
    active = "active", "The value is kept up-to-date"


class DeviceAttribute(DeviceProperty):
    """
    Model describing an attribute property of an IoT device.

    The attribute will add one field to the fiware instance:
    - {NameOfInstanceField}_{Name}, holds the value of the Iot device
    attribute: get_value()

    A DeviceAttribute can only belong to one field of one instance. Assigning
    it to multiple fields will result in an error.
    """

    model_config = ConfigDict(frozen=True, use_enum_values=True)
    attribute_type: DeviceAttributeType = Field(
        description="States if the attribute is read actively or lazy from "
        "the IoT Device into Fiware"
    )

    def get_value(self):
        """
        Retrieve the current value from the Iot Device

        Raises:
            Exception: If the DeviceAttribute was not yet saved to Fiware
        """
        return self._get_field_from_fiware(
            field_name=f"{self._instance_link.field_name}_{self.name}",
            required_type="StructuredValue",
        ).value

    def get_all_field_names(self, field_name: Optional[str] = None) -> List[str]:
        """
        Get all field names which this property creates in the fiware
        instance

        Args:
            field_name (str): Name of the field to which the attribute
                is/will be added. If none is provided, the linked field name
                is used
        """
        if field_name is None:
            field_name = self._instance_link.field_name
        return [f"{field_name}_{self.name}"]


class Field(BaseModel):
    """
    A Field corresponds to a CombinedRelation for a class from the vocabulary.
    It itself is a _set, that is enhanced with methods to provide validation
    of the values according to the rules stated in the vocabulary.

    The values of a field are unique and without order

    The fields of a class are predefined. A field can contain standard values
    on init
    """

    model_config = ConfigDict()

    name: str = Field(
        default="",
        description="Name of the Field, corresponds to the property name that "
        "it has in the SemanticClass",
    )

    _semantic_manager: "SemanticsManager"
    "Reference to the global SemanticsManager"

    _instance_identifier: InstanceIdentifier
    "Identifier of instance, that has this field as property"

    _set: Set = Field(
        default=set(),
        description="Internal set of the field, to which values are saved",
    )

    def __init__(self, name, semantic_manager):
        self._semantic_manager = semantic_manager
        super().__init__()
        self.name = name
        self._set = set()

    def is_valid(self) -> bool:
        """
        Check if the current state is valid -> Can be saved to Fiware
        """
        pass

    def build_context_attribute(self) -> NamedContextAttribute:
        """
        Convert the field to a NamedContextAttribute that can eb added to a
        ContextEntity

        Returns:
            NamedContextAttribute
        """
        pass

    def build_device_attributes(
        self,
    ) -> List[
        Union[
            iot.DeviceAttribute,
            iot.LazyDeviceAttribute,
            iot.StaticDeviceAttribute,
            iot.DeviceCommand,
        ]
    ]:
        """
        Convert the field to a DeviceAttribute that can eb added to a
        DeviceEntity

        Returns:
            List[Union[iot.DeviceAttribute,
                       iot.LazyDeviceAttribute,
                       iot.StaticDeviceAttribute,
                       iot.DeviceCommand]]
        """
        values = []
        for v in self.get_all_raw():
            if isinstance(v, BaseModel):
                values.append(v.model_dump())
            else:
                values.append(v)

        x = [
            iot.StaticDeviceAttribute(
                name=self.name,
                type=DataType.STRUCTUREDVALUE,
                value=values,
                entity_name=None,
                entity_type=None,
            )
        ]

        return x

    def __len__(self) -> int:
        """Get the number of values

        Returns:
            int
        """
        return len(self._set)

    def size(self) -> int:
        """Get the number of values

        Returns:
            int
        """
        return self.__len__()

    def remove(self, v):
        """
        Remove the given value from the field.

        Args:
            v, value that is in the field

        Raises:
            KeyError: if value not in field
        """
        self._set.remove(v)

    def add(self, v):
        """
        Add the value v to the field, duplicates are ignored (Set logic)

        Args:
            v (Any) Value to be added, of fitting type
        Raises:
            ValueError: if v is of invalid type
        """
        self._set.add(v)

    def update(self, values: Union[List, Set]):
        """
        Add all the values to the field, duplicates are ignored (Set logic)

        Args:
            values (Union[List, Set]): Values to be added, each of fitting type
        Raises:
            ValueError: if one value is of invalid type
        """
        for v in values:
            self.add(v)

    def set(self, values: List):
        """
        Set the values of the field equal to the given list

        Args:
            values: List of values fitting for the field

        Returns:
            None
        """
        self.clear()
        for v in values:
            self.add(v)

    def clear(self):
        """
        Remove all values of the field

        Returns:
            None
        """
        for v in self.get_all():
            self.remove(v)

    def __str__(self):
        """
        Get Field in a nice readable way

        Returns:
            str
        """
        result = f"Field: {self.name},\n\tvalues: ["
        values = self.get_all_raw()
        for value in values:
            result += f"{value}, "
        if len(values) > 0:
            result = result[:-2]
        return result

    def get_all_raw(self) -> Set:
        """
        Get all values of the field exactly as they are hold inside the
        internal list
        """
        return self._set

    def get_all(self) -> List:
        """
        Get all values of the field in usable form.
        Returns the set in List for as some values are not hashable in
        converted form.
        But the order is random

        Returns:
            List, unsorted
        """
        return [self._convert_value(v) for v in self._set]

    def _convert_value(self, v):
        """
        Converts the internal saved value v, to the type that should be returned
        """
        return v

    def _get_instance(self) -> "SemanticClass":
        """
        Get the instance object to which this field belongs
        """
        return self._semantic_manager.get_instance(self._instance_identifier)

    def get_field_names(self) -> List[str]:
        """
        Get the names of all fields this field will create in the Fiware entity.
        (DeviceProperties can create additional fields)

        Returns:
            List[str]
        """
        return [self.name]

    def values_to_json(self) -> List[str]:
        """
        Convert each value of the field to a json string

        Returns:
            List[str]
        """
        res = []
        for v in self.get_all_raw():
            if isinstance(v, BaseModel):
                res.append(v.model_dump_json())
            else:
                res.append(v)
        return res

    def __contains__(self, item) -> bool:
        """
        Overrides the magic "in" to test if a value/item is inside the field.

        Returns:
            bool
        """
        return item in self.get_all()

    def __iter__(self) -> Iterator[Any]:
        """
        Overrides the magic "in" to loop over the field values
        """
        return self.get_all().__iter__()


class DeviceField(Field):
    """
    A Field that represents a logical part of a device.
    Abstract Superclass
    """

    _internal_type: type = DeviceProperty
    """
    Type which is allowed to be stored in the field.
    Set in the subclasses, but has to be a subclass of DeviceProperty
    """

    def is_valid(self) -> bool:
        """
        Check if the current state is valid -> Can be saved to Fiware

        Returns:
            True, if all values are of type _internal_type
        """
        for value in self.get_all_raw():
            if not isinstance(value, self._internal_type):
                return False
        return True

    def name_check(self, v: _internal_type):
        """
        Executes name checks before value v is assigned to field values
        Each field name that v will add to the Fiware instance needs to be
        available

        Args:
            v (_internal_type): Value to be added to the field
        Raises:
            NameError: if a field name of v is not available
                       if a field name of v is blacklisted
                       if the name of v contains a forbidden character
        """
        taken_fields = self._get_instance().get_all_field_names()
        for name in v.get_all_field_names(field_name=self.name):
            if name in taken_fields:
                raise NameError(
                    f"The property can not be added to the field "
                    f"{self.name}, because the instance already"
                    f" posses a field with the name {name}"
                )
            if name in label_blacklist:
                raise NameError(
                    f"The property can not be added to the field "
                    f"{self.name}, because the name {name} is "
                    f"forbidden"
                )
            for c in name:
                if c not in label_char_whitelist:
                    raise NameError(
                        f"The property can not be added to the field "
                        f"{self.name}, because the name {name} "
                        f"contains the forbidden character {c}"
                    )

    def remove(self, v):
        """List function: Remove a values
        Makes the value available again to be added to other fields/instances
        """
        v._instance_link.instance_identifier = None
        v._instance_link.semantic_manager = None
        v._instance_link.field_name = None
        super(DeviceField, self).remove(v)

    def add(self, v):
        """List function: If checks pass , add value

        Args:
            v, value to add

        Raises:
            AssertionError, if v is of wrong type
            AssertionError, if v already belongs to a field
            NameError, if v has an invalid name

        Returns:
            None
        """

        # assert that the given value fulfills certain conditions
        assert isinstance(v, self._internal_type)
        assert isinstance(v, DeviceProperty)
        assert (
            v._instance_link.instance_identifier is None
        ), "DeviceProperty can only belong to one device instance"

        # test if name of v is valid, if not an error is raised
        self.name_check(v)

        # link attribute to field and instance
        v._instance_link.instance_identifier = self._instance_identifier
        v._instance_link.semantic_manager = self._semantic_manager
        v._instance_link.field_name = self.name

        super(DeviceField, self).add(v)

    def get_field_names(self) -> List[str]:
        """
        Get all names of fields that would be/are generated by this field in
        the fiware device_entity and its current values

        Returns:
            List[str]
        """
        names = super().get_field_names()
        for v in self.get_all_raw():
            names.extend(v.get_all_field_names())
        return names

    def build_context_attribute(self) -> NamedContextAttribute:
        """Export Field as NamedContextAttribute

        only needed when saving local state as json

        Returns:
            NamedContextAttribute
        """
        values = []
        for v in self.get_all_raw():
            if isinstance(v, BaseModel):
                values.append(v.model_dump())
            else:
                values.append(v)
        return NamedContextAttribute(name=self.name, value=values)


class CommandField(DeviceField):
    """
    A Field that holds commands that can be send to the device
    """

    _internal_type = Command

    def get_all_raw(self) -> Set[Command]:
        return super().get_all_raw()

    def get_all(self) -> List[Command]:
        return super().get_all()

    def __iter__(self) -> Iterator[Command]:
        return super().__iter__()

    def build_device_attributes(
        self,
    ) -> List[
        Union[
            iot.DeviceAttribute,
            iot.LazyDeviceAttribute,
            iot.StaticDeviceAttribute,
            iot.DeviceCommand,
        ]
    ]:
        attrs = super().build_device_attributes()
        for command in self.get_all_raw():
            attrs.append(
                iot.DeviceCommand(
                    name=command.name,
                )
            )
        return attrs


class DeviceAttributeField(DeviceField):
    """
    A Field that holds attributes of the device that can be referenced for
    live reading of the device
    """

    _internal_type = DeviceAttribute

    def get_all_raw(self) -> Set[DeviceAttribute]:
        return super().get_all_raw()

    def get_all(self) -> List[DeviceAttribute]:
        return super().get_all()

    def __iter__(self) -> Iterator[DeviceAttribute]:
        return super().__iter__()

    def build_device_attributes(
        self,
    ) -> List[
        Union[
            iot.DeviceAttribute,
            iot.LazyDeviceAttribute,
            iot.StaticDeviceAttribute,
            iot.DeviceCommand,
        ]
    ]:
        attrs = super().build_device_attributes()

        for attribute in self.get_all_raw():

            if attribute.attribute_type == DeviceAttributeType.active:
                attrs.append(
                    iot.DeviceAttribute(
                        object_id=attribute.name,
                        name=f"{self.name}_{attribute.name}",
                        type=DataType.STRUCTUREDVALUE,
                        entity_name=None,
                        entity_type=None,
                    )
                )
            else:
                attrs.append(
                    iot.LazyDeviceAttribute(
                        object_id=attribute.name,
                        name=f"{self.name}_{attribute.name}",
                        type=DataType.STRUCTUREDVALUE,
                        entity_name=None,
                        entity_type=None,
                    )
                )

        return attrs


class RuleField(Field):
    """
    A RuleField corresponds to a CombinedRelation for a class from the
    vocabulary.
    It itself is a list, that is enhanced with methods to provide validation
    of the values according to the rules stated in the vocabulary

    The fields of a class are predefined. A field can contain standard values
    on init
    """

    _rules: List[Tuple[str, List[List[str]]]]
    """rule formatted for machine readability """
    rule: str = pyd.Field(
        default="", description="rule formatted for human readability"
    )

    def __init__(self, rule, name, semantic_manager):
        self._semantic_manager = semantic_manager
        super().__init__(name, semantic_manager)
        self.rule = rule

    def is_valid(self) -> bool:
        """
        Check if the values present in this relationship fulfills the semantic
        rule.

        returns:
            bool
        """

        # true if all rules are fulfilled
        for [rule, fulfilled] in self.are_rules_fulfilled():
            if not fulfilled:
                return False
        return True

    def are_rules_fulfilled(self) -> List[Tuple[str, bool]]:
        """
        Check if the values present in this relationship fulfill the
        individual semantic rules.

        Returns:
            List[Tuple[str, bool]], [[readable_rule, fulfilled]]
        """

        # rule has form: (STATEMENT, [[a,b],[c],[a,..],..])
        # A value fulfills the rule if it is an instance of all the classes,
        #       datatype_catalogue listed in at least one innerlist
        # A field is fulfilled if a number of values fulfill the rule,
        #       the number is depending on the statement

        # The STATEMENTs and their according numbers are (STATEMENT|min|max):
        #       - only | len(values) | len(values)
        #       - some | 1 | len(values)
        #       - min n | n | len(values)
        #       - max n | 0 | n
        #       - range n,m | n | m

        res = []

        values = self.get_all()
        readable_rules = self.rule.split(",")
        rule_counter = 0

        # loop over all rules, if a rule is not fulfilled return False
        for rule in self._rules:
            # rule has form: (STATEMENT, [[a,b],[c],[a,..],..])
            statement: str = rule[0]
            outer_list: List[List] = rule[1]

            readable_rule = readable_rules[rule_counter].strip()
            rule_counter = rule_counter + 1

            # count how  many values fulfill this rule
            fulfilling_values = 0
            for v in values:

                # A value fulfills the rule if there exists an innerlist of
                # which the value is an instance of each value
                fulfilled = False
                for inner_list in outer_list:
                    counter = 0
                    for rule_value in inner_list:
                        if self._value_is_valid(v, rule_value):
                            counter += 1
                    if len(inner_list) == counter:
                        fulfilled = True

                if fulfilled:
                    fulfilling_values += 1

            # test if rule failed by evaluating the statement and the
            # number of fulfilling values
            if "min" in statement:
                number = int(statement.split("|")[1])
                if not fulfilling_values >= number:
                    res.append([readable_rule, False])
            elif "max" in statement:
                number = int(statement.split("|")[1])
                if not fulfilling_values <= number:
                    res.append([readable_rule, False])
            elif "exactly" in statement:
                number = int(statement.split("|")[1])
                if not fulfilling_values == number:
                    res.append([readable_rule, False])
            elif "some" in statement:
                if not fulfilling_values >= 1:
                    res.append([readable_rule, False])
            elif "only" in statement:
                if not fulfilling_values == len(values):
                    res.append([readable_rule, False])
            elif "value" in statement:
                if not fulfilling_values >= 1:
                    res.append([readable_rule, False])

            if len(res) == 0 or not (res[-1][0] == readable_rule):
                res.append([readable_rule, True])
        return res

    def _value_is_valid(self, value, rule_value) -> bool:
        """
        Test if a value of the field, fulfills a part of a rule

        Args:
            value: Value in field
            rule_value: Value from inner List of rules_

        Returns:
            bool, True if valid
        """
        pass

    def __str__(self):
        """
        Get Field in a nice readable way

        Returns:
            str
        """
        result = super(RuleField, self).__str__()
        result += f"],\n\trule: ({self.rule})"
        return result

    def _get_all_rule_type_names(self) -> Set[str]:
        """
        Returns the names all types mentioned in the field rule

        Returns:
            Set[str]
        """
        res = set()

        for rule in self._rules:
            statement: str = rule[0]
            outer_list: List[List] = rule[1]
            for inner_list in outer_list:
                for type_name in inner_list:
                    res.add(type_name)
        return res


class DataField(RuleField):
    """
    Field for CombinedDataRelation
    A Field that contains literal values: str, int, ...
    """

    def _value_is_valid(self, value, rule_value: str) -> bool:
        datatype = self._semantic_manager.get_datatype(rule_value)
        return datatype.value_is_valid(value)

    def build_context_attribute(self) -> NamedContextAttribute:
        return NamedContextAttribute(
            name=self.name,
            type=DataType.STRUCTUREDVALUE,
            value=[v for v in self.get_all_raw()],
        )

    def add(self, v):
        if isinstance(v, Enum):
            self._set.add(v.value)
        else:
            self._set.add(v)

    def __str__(self):
        return "Data" + super().__str__()

    def get_possible_enum_values(self) -> List[str]:
        """
        Get all enum values that are excepted for this field

        Returns:
            List[str]
        """
        enum_values = set()
        for type_name in self._get_all_rule_type_names():
            datatype = self._semantic_manager.get_datatype(type_name)
            if datatype.type == DatatypeType.enum:
                enum_values.update(datatype.enum_values)

        return sorted(enum_values)

    def get_all_possible_datatypes(self) -> List[Datatype]:
        """
        Get all Datatypes that are stated as allowed for this field.

        Returns:
            List[Datatype]
        """
        return [
            self._semantic_manager.get_datatype(type_name)
            for type_name in self._get_all_rule_type_names()
        ]


class RelationField(RuleField):
    """
    Field for CombinedObjectRelation
    A Field that contains links to other instances of SemanticClasses,
    or Individuals

    Internally this field only holds:
         - InstanceIdentifiers for SemanticClasses. If a value is accessed
             the corresponding instance is loaded form the local registry
             or hot loaded form Fiware
         - Names for Individuals. If a value is accessed a new object of
             that individual is returned (All instances are equal)
    """

    _rules: List[Tuple[str, List[List[Type]]]] = []
    inverse_of: List[str] = []
    """List of all field names which are inverse to this field.
    If an instance i1 is added to this field, the instance i2 belonging to this 
    field is added to all fields of i1 that are stated in this list by name"""

    def __init__(self, rule, name, semantic_manager, inverse_of=None):
        super().__init__(rule, name, semantic_manager)
        self.inverse_of = inverse_of

    def _value_is_valid(self, value, rule_value: type) -> bool:
        if isinstance(value, SemanticClass):
            return isinstance(value, rule_value)
        elif isinstance(value, SemanticIndividual):
            return value.is_instance_of_class(rule_value)
        else:
            return False

    def build_context_attribute(self) -> NamedContextAttribute:
        values = []
        for v in self.get_all_raw():
            if isinstance(v, InstanceIdentifier):
                values.append(v.model_dump())
            else:
                values.append(v)

        return NamedContextAttribute(
            name=self.name, type=DataType.RELATIONSHIP, value=values
        )

    def _convert_value(self, v):
        """
        Returns the internal holded objects as SemanticClass or
        SemanticIndividual
        """
        if isinstance(v, InstanceIdentifier):
            return self._semantic_manager.get_instance(v)
        elif isinstance(v, str):
            return self._semantic_manager.get_individual(v)

    def add(self, v: Union["SemanticClass", "SemanticIndividual"]):
        """see class description
        Raises:
            AttributeError: if value not an instance of 'SemanticClass' or
            'SemanticIndividual'
        """
        # self._uniqueness_check(v)

        if isinstance(v, SemanticClass):
            self._set.add(v.get_identifier())
            v.add_reference(self._instance_identifier, self.name)

            self._add_inverse(v)
        elif isinstance(v, SemanticIndividual):
            self._set.add(v.get_name())
        else:
            raise AttributeError(
                "Only instances of a SemanticClass or a "
                "SemanticIndividual can be given as value"
            )

    def remove(self, v):
        """see class description"""

        if isinstance(v, SemanticClass):
            identifier = v.get_identifier()
            assert identifier in self._set

            # delete reference
            if not self._semantic_manager.was_instance_deleted(identifier):
                v.remove_reference(self._instance_identifier, self.name)

            # delete value in field
            self._set.remove(identifier)

            # inverse of deletion
            if not self._semantic_manager.was_instance_deleted(identifier):
                # remove this instance in reverse fields
                if self.inverse_of is not None:
                    for inverse_field_name in self.inverse_of:
                        if inverse_field_name in v.get_all_field_names():
                            field = v.get_field_by_name(inverse_field_name)
                            if self._instance_identifier in field.get_all_raw():
                                field.remove(self._get_instance())
        elif isinstance(v, SemanticIndividual):
            self._set.remove(v.get_name())
        else:
            raise KeyError(
                f"v is neither of type SemanticIndividual nor SemanticClass but {type(v)}"
            )

    def _add_inverse(self, v: "SemanticClass"):
        """
        If a value is added to this field, and this field has an inverse
        logic field bound to it.
        It is tested if the added value posses that field.
        If yes the instance of this field is added to the inverse field of the
        added v
        """
        if self.inverse_of is not None:
            for inverse_field_name in self.inverse_of:
                if inverse_field_name in v.get_all_field_names():
                    field = v.get_field_by_name(inverse_field_name)
                    if self._instance_identifier not in field.get_all_raw():
                        field.add(self._get_instance())

    def __str__(self):
        """see class description"""
        return "Relation" + super().__str__()

    def __iter__(self) -> Iterator[Union["SemanticClass", "SemanticIndividual"]]:
        return super().__iter__()

    def get_all(self) -> List[Union["SemanticClass", "SemanticIndividual"]]:
        return super(RelationField, self).get_all()

    def get_all_raw(self) -> Set[Union[InstanceIdentifier, str]]:
        return super().get_all_raw()

    def get_all_possible_classes(
        self, include_subclasses: bool = False
    ) -> List[Type["SemanticClass"]]:
        """
        Get all SemanticClass types that are stated as allowed for this field.

        Args:
            include_subclasses (bool): If true all subclasses of target
                classes are also returned

        Returns:
            List[Type[SemanticClass]]
        """
        res = set()
        for class_name in self._get_all_rule_type_names():
            if class_name.__name__ in self._semantic_manager.class_catalogue:
                class_ = self._semantic_manager.get_class_by_name(class_name.__name__)
                res.add(class_)
                if include_subclasses:
                    res.update(class_.__subclasses__())

        return list(res)

    def get_all_possible_individuals(self) -> List["SemanticIndividual"]:
        """
        Get all SemanticIndividuals that are stated as allowed for this field.

        Returns:
            List['SemanticIndividual']
        """
        res = set()
        for name in self._get_all_rule_type_names():

            if name.__name__ not in self._semantic_manager.class_catalogue:
                res.add(self._semantic_manager.get_individual(name.__name__))
        return list(res)


class InstanceState(BaseModel):
    """State of instance that it had in Fiware on the moment of the last load
    Wrapped in an object to bypass the SemanticClass immutability
    """

    state: Optional[ContextEntity] = None


class SemanticMetadata(BaseModel):
    """
    Meta information about an semantic instance.
    A name and comment that can be used by the user to better identify the
    instance
    """

    model_config = ConfigDict(validate_assignment=True)
    name: str = pyd.Field(
        default="", description="Optional user-given name for the " "instance"
    )
    comment: str = pyd.Field(
        default="", description="Optional user-given comment for " "the instance"
    )


class SemanticClass(BaseModel):
    """
    A class representing a vocabulary/ontology class.
    A class has predefined fields
    Each instance of a class links to a unique Fiware ContextEntity (by
    Identifier)

    If a class is initiated it is first looked if this instance (equal over
    identifier) exists in the local registry. If yes that instance is returned

    If no, it is looked if this instance exists in Fiware, if yes it is
    loaded and returned, else a new instance of the class is initialised and
    returned
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, frozen=True)
    header: InstanceHeader = pyd.Field(
        description="Header of instance. Holds the information where the "
        "instance is saved in Fiware"
    )
    id: str = pyd.Field(
        description="Id of the instance, equal to Fiware ContextEntity Id",
        regex=FiwareRegex.standard.value,
    )

    old_state: InstanceState = pyd.Field(
        default=InstanceState(),
        description="State in Fiware the moment the instance was loaded "
        "in the local registry. Used when saving. "
        "Only the made changes are reflected",
    )

    references: Dict[InstanceIdentifier, List[str]] = pyd.Field(
        default={},
        description="references made to this instance in other instances "
        "RelationFields",
    )

    semantic_manager: BaseModel = pyd.Field(
        default=None,
        description="Pointer to the governing semantic_manager, "
        "vague type to prevent forward ref problems. "
        "But it will be of type 'SemanticsManager' in runtime",
    )

    metadata: SemanticMetadata = pyd.Field(
        default=SemanticMetadata(),
        description="Meta information about the instance. A name and comment "
        "that can be used by the user to better identify the "
        "instance",
    )

    def add_reference(self, identifier: InstanceIdentifier, relation_name: str):
        """
        Note that an instance references this instance in the relation

        Args:
            identifier (InstanceIdentifier): Identifier of the referencing
                                             instance
            relation_name (str): Field name in which the reference is taking
                                 place
        """
        if identifier not in self.references:
            self.references[identifier] = []
        self.references[identifier].append(relation_name)

    def remove_reference(self, identifier: InstanceIdentifier, relation_name: str):
        """
        Remove the note of reference

        Args:
           identifier (InstanceIdentifier): Identifier of the referencing
                                            instance
           relation_name (str): Field name in which the reference is taking
                                place
        """

        self.references[identifier].remove(relation_name)
        if len(self.references[identifier]) == 0:
            del self.references[identifier]

    def __new__(cls, *args, **kwargs):
        semantic_manager_ = kwargs["semantic_manager"]

        if "enforce_new" in kwargs:
            enforce_new = kwargs["enforce_new"]
        else:
            enforce_new = False

        if "identifier" in kwargs:
            instance_id = kwargs["identifier"].id
            header_ = kwargs["identifier"].header
            assert cls.__name__ == kwargs["identifier"].type
        else:
            instance_id = kwargs["id"] if "id" in kwargs else ""

            import re

            assert re.match(
                FiwareRegex.standard.value, instance_id
            ), "Invalid character in ID"

            header_ = (
                kwargs["header"]
                if "header" in kwargs
                else semantic_manager_.get_default_header()
            )

        if not instance_id == "" and not enforce_new:

            identifier = InstanceIdentifier(
                id=instance_id, type=cls.__name__, header=header_
            )

            if semantic_manager_.does_instance_exists(identifier=identifier):
                return semantic_manager_.load_instance(identifier=identifier)

        return super().__new__(cls)

    def __init__(self, *args, **kwargs):
        semantic_manager_ = kwargs["semantic_manager"]

        if "identifier" in kwargs:
            instance_id_ = kwargs["identifier"].id
            header_ = kwargs["identifier"].header
            assert self.get_type() == kwargs["identifier"].type
        else:
            instance_id_ = kwargs["id"] if "id" in kwargs else str(uuid.uuid4())
            header_ = (
                kwargs["header"]
                if "header" in kwargs
                else semantic_manager_.get_default_header()
            )

        # old_state_ = kwargs['old_state'] if 'old_state' in kwargs else None

        identifier_ = InstanceIdentifier(
            id=instance_id_,
            type=self.get_type(),
            header=header_,
        )

        if "enforce_new" in kwargs:
            enforce_new = kwargs["enforce_new"]
        else:
            enforce_new = False

        # test if this instance was taken out of the instance_registry instead
        # of being newly created. If yes abort __init__(), to prevent state
        # overwrite !
        if not enforce_new:
            if semantic_manager_.does_instance_exists(identifier_):
                return

        super().__init__(
            id=instance_id_,
            header=header_,
            semantic_manager=semantic_manager_,
            references={},
        )

        semantic_manager_.instance_registry.register(self)

    def is_valid(self) -> bool:
        """
        Test if instance is valid -> Is correctly defined and can be saved to
        Fiware

        Returns:
            bool
        """
        return self.are_rule_fields_valid()

    def are_rule_fields_valid(self) -> bool:
        """
        Test if all rule fields are valid

        Returns:
            bool, True if all valid
        """
        return len(self.get_invalid_rule_fields()) == 0

    def get_invalid_rule_fields(self) -> List[Field]:
        """
        Get all fields that are currently not valid

        Returns:
            List[Field]
        """
        return [f for f in self.get_rule_fields() if not f.is_valid()]

    def get_rule_fields(self) -> List[Field]:
        """
        Get all RuleFields of class

        Returns:
            List[Field]
        """
        res: List[Field] = self.get_relation_fields()
        res.extend(self.get_data_fields())
        return res

    def get_type(self) -> str:
        """
        Get _internal_type as used in Fiware, equal to class name

        Returns:
            str
        """
        return self._get_class_name()

    def _get_class_name(self) -> str:
        """
        Get name of class

        Returns:
            str
        """
        return type(self).__name__

    def delete(self, assert_no_references: bool = False):
        """
        Delete this instance.
        All references made to this instance by other instances will be removed
        On save_state it will also be deleted from Fiware

        Args:
            assert_no_references (bool): If True the instance is not deleted
            and an Error is raised, if some other instance references this
            instance.
        """

        if assert_no_references:
            assert len(self.references) == 0

        # remove all notes in other instances that they are referenced
        # clear all field data, this automatically handles the references
        for field in self.get_fields():
            field.clear()

        # remove all references in other instances
        for identifier, field_names in self.references.copy().items():
            for field_name in field_names:
                if not self.semantic_manager.was_instance_deleted(identifier):
                    instance = self.semantic_manager.get_instance(identifier)
                    instance.get_field_by_name(field_name).remove(self)

        self.semantic_manager.instance_registry.delete(self)

    def get_fields(self) -> List[RuleField]:
        """
        Get all fields of class

        Returns:
            List[Field]
        """
        fields: List[RuleField] = self.get_relation_fields()
        fields.extend(self.get_data_fields())
        return fields

    def get_relation_fields(self) -> List[RelationField]:
        """
        Get all RelationFields of class

        Returns:
           List[RelationField]
        """
        relationships = []
        for key, value in self.__dict__.items():
            if isinstance(value, RelationField):
                rel: RelationField = value
                relationships.append(rel)
        return relationships

    def get_data_fields(self) -> List[DataField]:
        """
        Get all DataFields of class

        Returns:
          List[DataField]
        """
        fields = []
        for key, value in self.__dict__.items():
            if isinstance(value, DataField):
                fields.append(value)
        return fields

    def get_relation_field_names(self) -> List[str]:
        """
        Get names of all RelationFields of class

        Returns:
            List[str]
        """
        return [f.name for f in self.get_relation_fields()]

    def get_data_field_names(self) -> List[str]:
        """
        Get names of all DataFields of class

        Returns:
            List[str]
        """
        return [f.name for f in self.get_data_fields()]

    def get_field_by_name(self, field_name: str) -> Field:
        """
        Get a field of class by its property name

        Raises:
            KeyError: If name does not belong to a field
        Returns:
            Field
        """
        for key, value in self.__dict__.items():
            if isinstance(value, Field):
                if value.name == field_name:
                    return value

        raise KeyError(
            f"{field_name} is not a valid Field for class " f"{self._get_class_name()}"
        )

    def _build_reference_dict(self) -> Dict:
        """
        Build the reference dict that is set as value in the context entity.
        We need to replace the . as it is a forbidden char for json keys,
        and IPs have .'s.

        The load  _context_entity_to_semantic_class loading teh object back
        again will reverse this swap

        Returns:
            Dict, with . replaced by ---
        """
        return {
            identifier.json().replace(".", "---"): value
            for (identifier, value) in self.references.items()
        }

    def build_context_entity(self) -> ContextEntity:
        """
        Convert the instance to a ContextEntity that contains all fields as
        NamedContextAttribute

        Returns:
            ContextEntity
        """
        entity = ContextEntity(id=self.id, type=self._get_class_name())

        for field in self.get_fields():
            entity.add_attributes([field.build_context_attribute()])

        reference_str_dict = self._build_reference_dict()

        # add meta attributes
        entity.add_attributes(
            [
                NamedContextAttribute(
                    name="referencedBy",
                    type=DataType.STRUCTUREDVALUE,
                    value=reference_str_dict,
                )
            ]
        )
        entity.add_attributes(
            [
                NamedContextAttribute(
                    name="metadata",
                    type=DataType.STRUCTUREDVALUE,
                    value=self.metadata.model_dump(),
                )
            ]
        )

        return entity

    def get_identifier(self) -> InstanceIdentifier:
        """
        Get identifier of instance, each instance's identifier is unique

        Returns:
            str
        """
        return InstanceIdentifier(id=self.id, type=self.get_type(), header=self.header)

    def get_all_field_names(self) -> List[str]:
        res = []
        for field in self.get_fields():
            res.extend(field.get_field_names())
        return res

    def __str__(self):
        return str(self.model_dump(exclude={"semantic_manager", "old_state"}))

    def __hash__(self):
        values = []
        for field in self.get_fields():
            values.extend((field.name, frozenset(field.get_all_raw())))

        ref_string = ""
        for ref in self.references.values():
            ref_string += f", {ref}"

        return hash(
            (
                self.id,
                self.header,
                self.metadata.name,
                self.metadata.comment,
                frozenset(self.references.keys()),
                ref_string,
                frozenset(values),
            )
        )


class SemanticDeviceClass(SemanticClass):
    """
    A class representing a vocabulary/ontology class.
    A class has predefined fields
    Each instance of a class links to a unique Fiware ContextDevice (by
    Identifier) and represents one IoT Device of the real world.

    If a class is initiated it is first looked if this instance (equal over
    identifier) exists in the local registry. If yes that instance is returned

    If no, it is looked if this instance exists in Fiware, if yes it is
    loaded and returned, else a new instance of the class is initialised and
    returned
    """

    device_settings: iot.DeviceSettings = pyd.Field(
        default=iot.DeviceSettings(),
        description="Settings configuring the communication with an IoT Device "
        "Wrapped in a model to bypass SemanticDeviceClass "
        "immutability",
    )

    def is_valid(self):
        """
        Test if instance is valid -> Is correctly defined and can be saved to
        Fiware

        Returns:
           bool
        """
        return super().is_valid() and self.are_device_settings_valid()

    def are_device_settings_valid(self):
        """
        Test if device settings are valid

        Returns:
             bool, True if endpoint and transport are not None
        """
        return self.device_settings.transport is not None

    def get_fields(self) -> List[Field]:
        """
        Get all fields of class

        Returns:
            List[str]
        """
        res = []
        for key, value in self.__dict__.items():
            if isinstance(value, Field):
                res.append(value)
        return res

    def get_command_fields(self) -> List[CommandField]:
        """
        Get all CommandFields of class

        Returns:
           List[CommandField]
        """
        commands = []
        for key, value in self.__dict__.items():
            if isinstance(value, CommandField):
                commands.append(value)
        return commands

    def get_device_attribute_fields(self) -> List[DeviceAttributeField]:
        """
        Get all DeviceAttributeField of class

        Returns:
          List[DeviceAttributeField]
        """
        fields = []
        for key, value in self.__dict__.items():
            if isinstance(value, DeviceAttributeField):
                fields.append(value)
        return fields

    def get_command_field_names(self) -> List[str]:
        """
        Get names of all CommandFields of class

        Returns:
            List[str]
        """
        return [f.name for f in self.get_command_fields()]

    def get_device_attribute_field_names(self) -> List[str]:
        """
        Get names of all DeviceAttributeFields of class

        Returns:
            List[str]
        """
        return [f.name for f in self.get_device_attribute_fields()]

    # needed when saving local state
    def build_context_entity(self) -> ContextEntity:
        entity = super(SemanticDeviceClass, self).build_context_entity()

        entity.add_attributes(
            [
                NamedContextAttribute(
                    name="deviceSettings",
                    type=DataType.STRUCTUREDVALUE,
                    value=self.device_settings.model_dump(),
                )
            ]
        )
        return entity

    def get_device_id(self) -> str:
        return f"{self.get_type()}|{self.id}"

    def build_context_device(self) -> iot.Device:
        """
        Convert the instance to a ContextEntity that contains all fields as
        NamedContextAttribute

        Returns:
            ContextEntity
        """
        device = iot.Device(
            device_id=self.get_device_id(),
            service=self.header.service,
            service_path=self.header.service_path,
            entity_name=f"{self.id}",
            entity_type=self._get_class_name(),
            apikey=self.device_settings.apikey,
            endpoint=self.device_settings.endpoint,
            protocol=self.device_settings.protocol,
            transport=self.device_settings.transport,
            timestamp=self.device_settings.timestamp,
            expressionLanguage=self.device_settings.expressionLanguage,
            ngsiVersion=self.header.ngsi_version,
        )

        for field in self.get_fields():
            for attr in field.build_device_attributes():
                device.add_attribute(attr)

        reference_str_dict = self._build_reference_dict()

        # add meta attributes
        device.add_attribute(
            iot.StaticDeviceAttribute(
                name="referencedBy",
                type=DataType.STRUCTUREDVALUE,
                value=reference_str_dict,
            )
        )
        device.add_attribute(
            iot.StaticDeviceAttribute(
                name="metadata",
                type=DataType.STRUCTUREDVALUE,
                value=self.metadata.model_dump(),
            )
        )
        device.add_attribute(
            iot.StaticDeviceAttribute(
                name="deviceSettings",
                type=DataType.STRUCTUREDVALUE,
                value=self.device_settings.model_dump(),
            )
        )

        return device


class SemanticIndividual(BaseModel):
    """
    A class representing a vocabulary/ontology Individual.
    A Individual has no fields and no values can be assigned

    It functions as some kind of enum value that can be inserted in
    RelationFields

    Each instance of an SemanticIndividual Class is equal
    """

    model_config = ConfigDict(frozen=True)
    _parent_classes: List[type] = pyd.Field(
        description="List of ontology parent classes needed to validate "
        "RelationFields"
    )

    def __eq__(self, other):
        """Each instance of an SemanticIndividual Class is equal"""
        return type(self) == type(other)

    def __str__(self):
        return type(self).__name__

    def get_name(self):
        """
        Get the name of the class

        Returns:
            str
        """
        return type(self).__name__

    def is_instance_of_class(self, class_: type) -> False:
        """
        Test if the individual is an instance of a class.

        Args:
            class_ (type): Class to be checked against

        Returns:
            bool, True if individual is of the searched class or one its parents
        """
        if isinstance(self, class_):
            return True
        for parent in self._parent_classes:
            # to test all subclasses correctly we need to instanciate parent
            # but it needs to be deleted directly again, as it is no real
            # instance that we want to keep in the local state
            parent = parent()
            is_instance = isinstance(parent, class_)
            parent.delete()
            if is_instance:
                return True
        return False



"""Vocabulary Models for CombinedRelations"""

from aenum import Enum

from filip.semantics.vocabulary import DataFieldType
from pydantic import BaseModel, Field
from typing import List, TYPE_CHECKING, Set
from . import Relation

if TYPE_CHECKING:
    from . import Vocabulary


class CombinedRelation(BaseModel):
    """
    Combines all relations of a class that share the same Property
    Represents one Field of a class. SHORT: CR

    It provides the common ground for the specialisations:
        CombinedObjectRelation, CombinedDataRelation
    """

    id: str = Field(description="Generated unique ID of the CR")
    relation_ids: List[str] = Field(
        default=[],
        description="List of all relations of the class that are "
        "bundled; have the same property",
    )
    property_iri: str = Field(
        description="IRI of the property, under which " "the relations are bundled"
    )
    class_iri: str = Field(
        description="IRI of the class the relations and " "this CR belongs to"
    )

    def get_relations(self, vocabulary: "Vocabulary") -> List[Relation]:
        result = []
        for id in self.relation_ids:
            result.append(vocabulary.get_relation_by_id(id))

        return result

    def get_property_label(self, vocabulary: "Vocabulary") -> str:
        """Get the label of the Property. Overwritten by children

        Args:
            vocabulary (Vocabulary): Vocabulary of the project

        Returns:
            str
        """
        return ""

    def get_all_targetstatements_as_string(self, vocabulary: "Vocabulary") -> str:
        """
        Get a string stating all conditions(target statement) that need to
        be fulfilled, so that this CR is fulfilled

        Args:
            vocabulary (Vocabulary): Vocabulary of the project

        Returns:
            str
        """
        res = ""
        for relation in self.get_relations(vocabulary=vocabulary):
            res = res + relation.to_string(vocabulary) + ", "

        return res[:-2]

    def get_all_target_iris(self, vocabulary: "Vocabulary") -> Set[str]:
        """Get all iris of referenced targets

        Args:
            vocabulary (Vocabulary): Vocabulary of the project

        Returns:
            set(str)
        """
        iris = set()

        for relation_id in self.relation_ids:
            relation = vocabulary.get_relation_by_id(relation_id)
            iris.update(relation.get_all_target_iris())
        return iris

    def get_all_target_labels(self, vocabulary: "Vocabulary") -> Set[str]:
        """Get all labels of referenced targets

        Args:
            vocabulary (Vocabulary): Vocabulary of the project

        Returns:
            set(str)
        """
        return {
            vocabulary.get_label_for_entity_iri(iri)
            for iri in self.get_all_target_iris(vocabulary)
        }

    def export_rule(self, vocabulary: "Vocabulary", stringify_fields: bool) -> str:
        """Get the rule as string

        Args:
           vocabulary (Vocabulary): Vocabulary of the project
           stringify_fields (bool): If true, all string delimieters will be
                removed

        Returns:
           str
        """

        rules = [
            vocabulary.get_relation_by_id(id).export_rule(vocabulary)
            for id in self.relation_ids
        ]
        if stringify_fields:
            return str(rules).replace('"', "")
        else:
            return str(rules).replace("'", "").replace('"', "'")


class CombinedDataRelation(CombinedRelation):
    """
    Combines all data relations of a class that share the same DataProperty
    Represents one Data Field of a class
    """

    def get_property_label(self, vocabulary: "Vocabulary") -> str:
        """Get the label of the DataProperty

        Args:
            vocabulary (Vocabulary): Vocabulary of the project

        Returns:
            str
        """
        return vocabulary.get_data_property(self.property_iri).get_label()

    def get_possible_enum_target_values(self, vocabulary: "Vocabulary") -> List[str]:
        """Get all enum values that are allowed as values for this Data field

        Args:
            vocabulary (Vocabulary): Vocabulary of the project

        Returns:
            List[str]
        """

        enum_values = set()

        for relation in self.get_relations(vocabulary):
            for value in relation.get_possible_enum_target_values(vocabulary):
                enum_values.add(value)

        return sorted(list(enum_values))

    def get_field_type(self, vocabulary: "Vocabulary") -> DataFieldType:
        """Get type of CDR (command, devicedata , simple)

         Args:
            vocabulary (Vocabulary): Vocabulary of the project

        Returns:
            DataFieldType
        """
        property = vocabulary.get_data_property(self.property_iri)
        return property.field_type

    def is_device_relation(self, vocabulary: "Vocabulary") -> bool:
        """Test if the CDR is a device property(command, or readings)

        Args:
            vocabulary (Vocabulary): Vocabulary of the project

        Returns:
            bool
        """
        return not self.get_field_type(vocabulary) == DataFieldType.simple


class CombinedObjectRelation(CombinedRelation):
    """
    Combines all object relations of a class that share the same ObjectProperty
    Represents one Relation Field of a class
    """

    def get_all_possible_target_class_iris(self, vocabulary) -> List[str]:
        """Get all iris that are valid values for this cor

        Args:
            vocabulary (Vocabulary): Vocabulary of the project

        Returns:
            List[str]
        """
        from . import Vocabulary

        assert isinstance(vocabulary, Vocabulary)

        relations = self.get_relations(vocabulary)
        result_set = set()
        for relation in relations:
            result_set.update(relation.get_all_possible_target_class_iris(vocabulary))

        return list(result_set)

    def get_property_label(self, vocabulary: "Vocabulary") -> str:
        """Get the label of the ObjectProperty

        Args:
            vocabulary (Vocabulary): Vocabulary of the project

        Returns:
            str
        """
        return vocabulary.get_object_property(self.property_iri).get_label()

    def get_inverse_of_labels(self, vocabulary: "Vocabulary") -> List[str]:
        """Get the labels of the inverse_of properties of this COR

        Args:
            vocabulary (Vocabulary): Vocabulary of the project

        Returns:
            List[str]
        """
        property = vocabulary.get_object_property(self.property_iri)
        return [
            vocabulary.get_entity_by_iri(iri).label
            for iri in property.inverse_property_iris
        ]



"""Main Vocabulary Model, and Submodels"""

import operator
from enum import Enum

from pydantic import BaseModel, Field
from . import *
from typing import List, Dict, Union, Optional, Tuple

from ...models.base import LogLevel


class LabelSummary(BaseModel):
    """
    Model holding all information for label conflicts in a vocabulary
    """

    class_label_duplicates: Dict[str, List[Entity]] = Field(
        description="All Labels that are used more than once for class_names "
        "on export."
        "Key: Label, Values: List of entities with key label"
    )
    field_label_duplicates: Dict[str, List[Entity]] = Field(
        description="All Labels that are used more than once for property_names"
        "on export."
        "Key: Label, Values: List of entities with key label"
    )
    datatype_label_duplicates: Dict[str, List[Entity]] = Field(
        description="All Labels that are used more than once for datatype "
        "on export."
        "Key: Label, Values: List of entities with key label"
    )

    blacklisted_labels: List[Tuple[str, Entity]] = Field(
        description="All Labels that are blacklisted, "
        "Tuple(Label, Entity with label)"
    )
    labels_with_illegal_chars: List[Tuple[str, Entity]] = Field(
        description="All Labels that contain illegal characters, "
        "Tuple(Label, Entity with label)"
    )

    def is_valid(self) -> bool:
        """Test if Labels are valid

        Returns:
            bool, True if no entries exist
        """
        return (
            len(self.class_label_duplicates) == 0
            and len(self.field_label_duplicates) == 0
            and len(self.datatype_label_duplicates) == 0
            and len(self.blacklisted_labels) == 0
            and len(self.labels_with_illegal_chars) == 0
        )

    def __str__(self):
        res = ""

        def print_collection(collection):
            sub_res = ""
            for key, values in collection.items():
                sub_res += f"\t{key}: "
                for v in values:
                    sub_res += f" \n\t\t{v.iri}"
                sub_res += "\n"

            if len(collection) == 0:
                sub_res += "\t/\n"
            return sub_res

        def print_list(collection):
            sub_res = ""
            for key, value in collection:
                sub_res += f"\t{key}: \t {value.iri}"
                sub_res += "\n"
            if len(collection) == 0:
                sub_res += "\t/\n"
            return sub_res

        res += "class_label_duplicates:\n"
        res += print_collection(self.class_label_duplicates)
        res += "field_label_duplicates:\n"
        res += print_collection(self.field_label_duplicates)
        res += "datatype_label_duplicates:\n"
        res += print_collection(self.datatype_label_duplicates)
        res += "blacklisted_labels:\n"
        res += print_list(self.blacklisted_labels)
        res += "labels_with_illegal_chars:\n"
        res += print_list(self.labels_with_illegal_chars)
        return res


class IdType(str, Enum):
    """Type of object that is referenced by an id/iri"""

    class_ = "Class"
    object_property = "Object Property"
    data_property = "Data Property"
    datatype = "Datatype"
    relation = "Relation"
    combined_relation = "Combined Relation"
    individual = "Individual"
    source = "Source"


class VocabularySettings(BaseModel):
    """
    Settings that state how labels of ontology entities should be
    automatically converted on parsing
    """

    pascal_case_class_labels: bool = Field(
        default=True,
        description="If true, convert all class labels given in the ontologies "
        "to PascalCase",
    )
    pascal_case_individual_labels: bool = Field(
        default=True,
        description="If true, convert all labels of individuals given in the "
        "ontologies to PascalCase",
    )
    camel_case_property_labels: bool = Field(
        default=True,
        description="If true, convert all labels of properties given in the "
        "ontologies to camelCase",
    )
    camel_case_datatype_labels: bool = Field(
        default=True,
        description="If true, convert all labels of datatypes given in the "
        "ontologies to camelCase",
    )
    pascal_case_datatype_enum_labels: bool = Field(
        default=True,
        description="If true, convert all values of enum datatypes given in "
        "the to PascalCase",
    )


class Vocabulary(BaseModel):
    """
    Semantic Vocabulary of a project

    This class holds all objects in a vocabulary as central unit.
    These objects can be accessed with the according ids/iris

    The vocabulary consists out of multiple sources, that each contribute
    objects

    From the vocabulary nothing should be added or deleted manually. The sources
    are added and removed through the respective methods. Everything else should
    be used as READ-ONLY
    """

    classes: Dict[str, Class] = Field(
        default={}, description="Classes of the vocabulary. Key: class_iri"
    )
    object_properties: Dict[str, ObjectProperty] = Field(
        default={},
        description="ObjectProperties of the vocabulary. " "Key: object_property_iri",
    )
    data_properties: Dict[str, DataProperty] = Field(
        default={},
        description="DataProperties of the vocabulary. Key: data_property_iri",
    )
    datatypes: Dict[str, Datatype] = Field(
        default={}, description="Datatypes of the vocabulary. Key: datatype_iri"
    )
    individuals: Dict[str, Individual] = Field(
        default={}, description="Individuals in the vocabulary. Key: individual_iri"
    )

    relations: Dict[str, Relation] = Field(
        default={},
        description="Relations of classes in the vocabulary. Key: relation_id",
    )
    combined_object_relations: Dict[str, CombinedObjectRelation] = Field(
        default={},
        description="CombinedObjectRelations of classes in the vocabulary."
        " Key: combined_relation_id",
    )
    combined_data_relations: Dict[str, CombinedDataRelation] = Field(
        default={},
        description="CombinedDataRelations of classes in the vocabulary."
        "Key: combined_data_id",
    )

    sources: Dict[str, Source] = Field(
        default={}, description="Sources of the vocabulary. Key: source_id"
    )

    id_types: Dict[str, IdType] = Field(
        default={},
        description="Maps all entity iris and (combined)relations to their "
        "Entity/Object type, to speed up lookups",
    )

    original_label_summary: Optional[LabelSummary] = Field(
        default=None,
        description="Original label after parsing, before the user made " "changes",
    )

    settings: VocabularySettings = Field(
        default=VocabularySettings(),
        description="Settings how to auto transform the entity labels",
    )

    def get_type_of_id(self, id: str) -> Union[IdType, None]:
        """Get the type (class, relation,...) of an iri/id

        Args:
            id (str): id/iri of which the type should be returned

        Returns:
            IdType or None if id/iri not registered
        """
        try:
            return self.id_types[id]
        except KeyError:
            return None

    def get_class_by_iri(self, class_iri: str) -> Union[Class, None]:
        """Get the class belonging to the class_iri

        Args:
            class_iri (str): iri

        Returns:
            Class or None if iri not a registered class
        """
        return self.classes.get(class_iri)

    def get_object_property(self, obj_prop_iri: str) -> ObjectProperty:
        """Get the object property beloning to the iri

        Args:
            obj_prop_iri (str): iri

        Returns:
            ObjectProperty

        Raises:
            KeyError: if not a registered obj property

        """
        return self.object_properties[obj_prop_iri]

    def get_data_property(self, data_prop_iri: str) -> DataProperty:
        """Get the data property belonging to the iri

        Args:
            data_prop_iri (str): iri

        Returns:
            DataProperty

        Raises:
            KeyError: if not a registered obj property

        """
        return self.data_properties[data_prop_iri]

    def iri_is_predefined_datatype(self, iri: str) -> bool:
        """Test if an iri belongs to a predefined datatype

        Args:
            iri (str): Iri to test

        Returns:
            bool
        """
        if iri not in self.id_types.keys():
            return False
        if self.id_types[iri] is IdType.datatype:
            return self.get_datatype(iri).predefined

    def get_datatype(self, datatype_iri: str) -> Datatype:
        """Get the datatype belonging to the iri

        Args:
            datatype_iri (str): iri

        Returns:
            DataType

        Raises:
            KeyError: if not a registered datatype
        """
        return self.datatypes[datatype_iri]

    def get_individual(self, individual_iri: str) -> Individual:
        """Get the individual belonging to the iri

        Args:
            individual_iri (str): iri

        Returns:
             Individual

        Raises:
            KeyError: if not a registered  Individual
        """
        return self.individuals[individual_iri]

    def get_all_individuals_of_class(self, class_iri: str) -> List[Individual]:
        """Get all individual that have the class_iri as parent

        Args:
             class_iri (str): iri of parent class

        Returns:
             List[Individual]
        """
        result = []
        for individual in self.individuals.values():
            if class_iri in individual.parent_class_iris:
                result.append(individual)
        return result

    def is_id_from_individual(self, id: str) -> bool:
        """
        Test if an id is from an Individual. Used to distinguish between
        instances and individuals

        Args:
            id (str): id

        Returns:
            bool
        """
        try:
            return self.get_type_of_id(id) == IdType.individual
        except:
            return False

    def get_classes(self) -> List[Class]:
        """
        Get all classes in this vocabulary

        Returns:
            List[Class]
        """
        return list(self.classes.values())

    def get_classes_sorted_by_label(self) -> List[Class]:
        """Get all classes sorted by their labels

        Returns:
           List[Class]: sorted classes, ascending
        """
        return sorted(
            self.classes.values(), key=operator.methodcaller("get_label"), reverse=False
        )

    def get_entity_list_sorted_by_label(self, list: List[Entity]) -> List[Entity]:
        """Sort a given entity list by their labels

        Args:
            list (List[Entity]) : entities to be sorted
        Returns:
            List[Entity]: sorted list
        """
        return sorted(list, key=operator.methodcaller("get_label"), reverse=False)

    def get_object_properties_sorted_by_label(self) -> List[ObjectProperty]:
        """Get all object properties of the vocabulary sorted by their labels

        Returns:
            List[ObjectProperty], sorted by ascending labels
        """
        return sorted(
            self.object_properties.values(),
            key=operator.methodcaller("get_label"),
            reverse=False,
        )

    def get_data_properties_sorted_by_label(self) -> List[DataProperty]:
        """Get all data properties of the vocabulary sorted by their labels

        Returns:
            List[DataProperty], sorted by ascending labels
        """
        return sorted(
            self.data_properties.values(),
            key=operator.methodcaller("get_label"),
            reverse=False,
        )

    def get_individuals_sorted_by_label(self) -> List[Individual]:
        """Get all individuals of the vocabulary sorted by their labels

        Returns:
            List[Individual], sorted by ascending labels
        """
        return sorted(
            self.individuals.values(),
            key=operator.methodcaller("get_label"),
            reverse=False,
        )

    def get_datatypes_sorted_by_label(self) -> List[Datatype]:
        """Get all datatypes of the vocabulary sorted by their labels

        Returns:
            List[Datatype], sorted by ascending labels
        """
        return sorted(
            self.datatypes.values(),
            key=operator.methodcaller("get_label"),
            reverse=False,
        )

    def get_relation_by_id(self, id: str) -> Relation:
        """Get Relation by relation id

        Args:
            id (str): relation_id

        Returns:
            Relation

        Raises:
            KeyError: if id not registered as relation
        """
        return self.relations[id]

    def get_combined_relation_by_id(self, id: str) -> CombinedRelation:
        """Get CombinedRelation by id

        Args:
            id (str): combined_relation_id

        Returns:
            CombinedRelation

        Raises:
            KeyError: if id not registered as CombinedObjectRelation
                        or CombinedDataRelation
        """
        if id in self.combined_object_relations:
            return self.combined_object_relations[id]
        else:
            return self.combined_data_relations[id]

    def get_combined_data_relation_by_id(self, id: str) -> CombinedDataRelation:
        """Get CombinedDataRelation by id

        Args:
            id (str): combined_relation_id

        Returns:
            CombinedDataRelation

        Raises:
            KeyError: if id not registered as CombinedDataRelation
        """
        return self.combined_data_relations[id]

    def get_combined_object_relation_by_id(self, id: str) -> CombinedObjectRelation:
        """Get CombinedObjectRelation by id

        Args:
            id (str): combined_relation_id

        Returns:
            CombinedObjectRelation

        Raises:
            KeyError: if id not registered as CombinedObjectRelation
        """
        return self.combined_object_relations[id]

    def get_source(self, source_id: str) -> Source:
        """Get the source with the given id

        Args:
            source_id (str): id

        Returns:
            Source

        Raises:
            KeyError: if source_id is not registered
        """
        return self.sources[source_id]

    def get_source_list(self) -> List[Source]:
        """Get all source objects of the vocabulary as list

        Returns:
            List[Source]
        """
        res = []
        for iri in self.sources:
            res.append(self.sources[iri])
        return res

    def has_source(self, source_id: str) -> bool:
        """Test if the vocabulary contains a source with the given id

        Args:
            source_id (str): id to test

        Returns:
            bool
        """
        return source_id in self.sources

    def is_id_of_type(self, id: str, type: IdType) -> bool:
        """Test if an iri/id is of a given type

        Args:
            id (str): id to test
            type (str): Type to test against

        Returns:
            bool

        Raises:
            KeyError: if id not registered
        """
        return self.id_types[id] == type

    def get_label_for_entity_iri(self, iri: str) -> str:
        """Get the label of the entity with the given iri
        Fast efficient methode

        Args:
            iri (str)

        Returns:
            str, "" if iri does not belong to an entity
        """

        entity = self.get_entity_by_iri(iri)
        if entity is not None:
            return entity.get_label()
        else:
            return ""

    @staticmethod
    def get_base_out_of_iri(iri: str):
        """Extract out of a given iri the base aka ontology name

        Args:
            iri (str), iri to extract

        Returns:
            str, base of iri
        """

        if "#" in iri:
            index = iri.find("#")
            return iri[:index]
        else:
            # for example if uri looks like:
            # http://webprotege.stanford.edu/RDwpQ8vbi7HaApq8VoqJUXH
            index = iri.rfind("/")
            return iri[:index]

    def get_entity_by_iri(self, iri: str) -> Union[None, Entity]:
        """Get the entity with the given iri
        Fast efficient methode

        Args:
            iri (str)

        Returns:
            Entity or None if iri does not belong to an Entity
        """
        if iri not in self.id_types:
            return None
        else:
            id_type = self.get_type_of_id(iri)

            if id_type == IdType.individual:
                return self.get_individual(iri)
            if id_type == IdType.class_:
                return self.get_class_by_iri(iri)
            if id_type == IdType.datatype:
                return self.get_datatype(iri)
            if id_type == IdType.object_property:
                return self.get_object_property(iri)
            if id_type == IdType.data_property:
                return self.get_data_property(iri)
            else:
                return None

    def is_iri_registered(self, iri: str) -> bool:
        """Test if iri/id is registered (Entities or (Combined)relations)

        Args:
            iri (str): iri to test

        Returns:
            bool
        """
        return iri in self.id_types

    def set_label_for_entity(self, iri: str, label: str):
        """Set a userset label for the given entity

        Args:
            iri (str): entity iri
            label (str): new label

        Returns:
            None
        """
        entity = self.get_entity_by_iri(iri)

        if entity.get_original_label() == label:
            entity.user_set_label = ""
        elif label == "":
            entity.user_set_label = ""

        else:
            entity.user_set_label = label

    def get_all_entities(self) -> List[Entity]:
        """Get all registered Entities

        Returns:
            List[Entity]
        """
        lists = [
            self.classes.values(),
            self.object_properties.values(),
            self.data_properties.values(),
            self.datatypes.values(),
            self.individuals.values(),
        ]

        res = []
        for l in lists:
            res.extend(l)
        return res

    def get_enum_dataytypes(self) -> Dict[str, Datatype]:
        """Get all datatypes of vocabularies that are of type ENUM

        Returns:
            Dict[str, Datatype], {datatype.iri: Datatype}
        """
        return {
            datatype.iri: datatype
            for datatype in self.datatypes.values()
            if len(datatype.enum_values) > 0 and not datatype.predefined
        }



"""
Implementation of UN/CEFACT units

We creating the data set of UNECE units from here.
"https://github.com/datasets/unece-units-of-measure"
It downloads the data and stores it in external resources if not
already present. For additional information on UNECE an the current state of
tables visit this website:
https://unece.org/trade/cefact/UNLOCODE-Download
https://unece.org/trade/uncefact/cl-recommendations
"""

import json
import logging
import pandas as pd
from functools import lru_cache
from rapidfuzz import process
from typing import Any, Dict, List, Optional, Union
from typing_extensions import Literal
from pydantic import field_validator, model_validator, ConfigDict, BaseModel, Field
from filip.models.base import NgsiVersion, DataType
from filip.utils.data import load_datapackage


logger = logging.getLogger(name=__name__)


@lru_cache()
def load_units() -> pd.DataFrame:
    """
    Loads data package from public repo if not already downloaded.
    This function will be cached for fast accessing the data set.
    Returns:
        Cleaned dataset containing all unit data
    """
    units = load_datapackage(
        url="https://github.com/datasets/unece-units-of-measure",
        package_name="unece-units",
    )["units_of_measure"]
    # remove deprecated entries
    units = units.loc[
        ((units.Status.str.casefold() != "x") & (units.Status.str.casefold() != "d"))
    ]
    return units


class UnitCode(BaseModel):
    """
    The unit of measurement given using the UN/CEFACT Common Code (3 characters)
    or a URL. Other codes than the UN/CEFACT Common Code may be used with a
    prefix followed by a colon.
    https://schema.org/unitCode

    Note:
        Currently we only support the UN/CEFACT Common Codes
    """

    type: DataType = Field(
        default=DataType.TEXT,
        # const=True,
        description="Data type",
    )
    value: str = Field(
        ...,
        title="Code of unit ",
        description="UN/CEFACT Common Code (3 characters)",
        min_length=2,
        max_length=3,
    )

    @field_validator("value")
    @classmethod
    def validate_code(cls, value):
        units = load_units()
        if len(units.loc[units.CommonCode == value.upper()]) == 1:
            return value
        raise KeyError("Code does not exist or is deprecated! '%s'", value)


class UnitText(BaseModel):
    """
    A string or text indicating the unit of measurement. Useful if you cannot
    provide a standard unit code for unitCode.
    https://schema.org/unitText

    Note:
        We use the names of units of measurements from UN/CEFACT for validation
    """

    type: DataType = Field(
        default=DataType.TEXT,
        # const=True,
        description="Data type",
    )
    value: str = Field(
        ...,
        title="Name of unit of measurement",
        description="Verbose name of a unit using British "
        "spelling in singular form, "
        "e.g. 'newton second per metre'",
    )

    @field_validator("value")
    @classmethod
    def validate_text(cls, value):
        units = load_units()

        if len(units.loc[(units.Name.str.casefold() == value.casefold())]) >= 1:
            return value
        names = units.Name.tolist()
        suggestions = [
            item[0]
            for item in process.extract(
                query=value.casefold(), choices=names, score_cutoff=50, limit=5
            )
        ]
        raise ValueError(
            f"Invalid 'name' for unit! '{value}' \n "
            f"Did you mean one of the following? \n "
            f"{suggestions}"
        )


class Unit(BaseModel):
    """
    Model for a unit definition
    """

    model_config = ConfigDict(extra="ignore", populate_by_name=True)
    _ngsi_version: Literal[NgsiVersion.v2] = NgsiVersion.v2
    name: Optional[Union[str, UnitText]] = Field(
        alias="unitText",
        default=None,
        description="A string or text indicating the unit of measurement",
    )
    code: Optional[Union[str, UnitCode]] = Field(
        alias="unitCode",
        default=None,
        description="The unit of measurement given using the UN/CEFACT "
        "Common Code (3 characters)",
    )
    description: Optional[str] = Field(
        default=None,
        alias="unitDescription",
        description="Verbose description of unit",
        max_length=350,
    )
    symbol: Optional[str] = Field(
        default=None,
        alias="unitSymbol",
        description="The symbol used to represent the unit of measure as "
        "in ISO 31 / 80000.",
    )
    conversion_factor: Optional[str] = Field(
        default=None,
        alias="unitConversionFactor",
        description="The value used to convert units to the equivalent SI "
        "unit when applicable.",
    )

    @model_validator(mode="before")
    @classmethod
    def check_consistency(cls, values):
        """
        Validate and auto complete unit data based on the UN/CEFACT data
        Args:
            values (dict): Values of a all data fields

        Returns:
            values (dict): Validated data
        """
        units = load_units()
        name = values.get("name")
        code = values.get("code")

        if isinstance(name, dict):
            name = UnitText.model_validate(name)
        if isinstance(code, UnitCode):
            code = code.value
        if isinstance(name, UnitText):
            name = name.value

        if code and name:
            idx = units.index[((units.CommonCode == code) & (units.Name == name))]
            if idx.empty:
                raise ValueError(
                    "Invalid combination of 'code' and 'name': ", code, name
                )
        elif code:
            idx = units.index[(units.CommonCode == code)]
            if idx.empty:
                raise ValueError("Invalid 'code': ", code)
        elif name:
            idx = units.index[(units.Name == name)]
            if idx.empty:
                names = units.Name.tolist()
                suggestions = [
                    item[0]
                    for item in process.extract(
                        query=name.casefold(), choices=names, score_cutoff=50, limit=5
                    )
                ]

                raise ValueError(
                    f"Invalid 'name' for unit! '{name}' \n "
                    f"Did you mean one of the following? \n "
                    f"{suggestions}"
                )
        else:
            raise AssertionError("'name' or 'code' must be  provided!")

        values["code"] = UnitCode(value=units.CommonCode[idx[0]]).value
        values["name"] = UnitText(value=units.Name[idx[0]]).value
        values["symbol"] = units.Symbol[idx[0]]
        values["conversion_factor"] = units.ConversionFactor[idx[0]]
        if not values.get("description"):
            values["description"] = units.Description[idx[0]]
        return values


class Units:
    """
    Class for easy accessing the data set of UNECE units from here.
    "https://github.com/datasets/unece-units-of-measure"
    """

    units = load_units()

    def __getattr__(self, item):
        """
        Return unit as attribute by name or code.
        Notes:
            Underscores will be substituted with whitespaces
        Args:
            item: if len(row) == 0:

        Returns:
            Unit
        """
        item = item.casefold().replace("_", " ")
        return self.__getitem__(item)

    @property
    def quantities(self):
        """
        Get list of units ordered by measured quantities
        Returns:
            list of units ordered by measured quantities
        """
        raise NotImplementedError(
            "The used dataset does currently not "
            "contain the information about quantity"
        )

    def __getitem__(self, item: str) -> Unit:
        """
        Get unit by name or code

        Args:
            item (str): name or code

        Returns:
            Unit
        """
        idx = self.units.index[
            (
                (self.units.CommonCode == item.upper())
                | (self.units.Name.str.casefold() == item.casefold())
            )
        ]
        if idx.empty:
            names = self.units.Name.tolist()
            suggestions = [
                item[0]
                for item in process.extract(
                    query=item.casefold(), choices=names, score_cutoff=50, limit=5
                )
            ]
            raise ValueError(
                f"Invalid 'name' for unit! '{item}' \n "
                f"Did you mean one of the following? \n "
                f"{suggestions}"
            )

        return Unit(code=self.units.CommonCode[idx[0]])

    @classmethod
    def keys(cls, by_code: bool = False) -> List[str]:
        """
        Returns list of all unit names or codes

        Args:
            by_code (bool): if 'True' the keys will contain the unit codes
                instead of their names.

        Returns:
            List[str] containing the names or list
        """
        if by_code:
            return cls.units.CommonCode.to_list()
        return cls.units.Name.to_list()

    @property
    def names(self) -> List[str]:
        """
        Returns list of all unit names

        Returns:
            List[str] containing the names or list
        """
        return self.keys()

    @property
    def codes(self) -> List[str]:
        """
        Returns list of all unit codes

        Returns:
            List[str] containing the codes
        """
        return self.keys(by_code=True)

    def values(self) -> List[Unit]:
        """
        Get list of all units

        Returns:
            List[Unit] containing all units
        """

        return [Unit(code=code) for code in self.units.CommonCode]

    def get(self, item: str, default: Any = None):
        """
        Get unit by name or by code

        Args:
            item (str): name or code of unit
            default (Any): Default value to return if unit does not exist.
        Returns:
            Unit
        """
        try:
            return self.__getitem__(item)
        except KeyError:
            return default


def validate_unit_data(data: Dict) -> Dict:
    """
    Validator for unit objects
    Args:
        data (Dict): Dictionary containing the metadata of an object

    Returns:
        Validated dictionary of metadata
    """
    _unit_models = {"unit": Unit, "unitText": UnitText, "unitCode": UnitCode}
    for modelname, model in _unit_models.items():
        if data.get("name", "").casefold() == modelname.casefold():
            if data.get("name", "").casefold() == "unit":
                data["type"] = "Unit"
                data["value"] = model.model_validate(data["value"])
                # data["value"] = model.parse_obj(data["value"])
                return data
            else:
                data.update(model.model_validate(data).model_dump())
                # data.update(model.parse_obj(data).dict())
                return data
    raise ValueError(f"Invalid unit data found: \n " f"{json.dumps(data, indent=2)}")
